

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import re
import numpy as np
import warnings
from functools import partial

from mmgroup.structures.abstract_group import singleton
from mmgroup.structures.abstract_mm_group import AbstractMMGroupWord
from mmgroup.structures.abstract_mm_group import AbstractMMGroup
from mmgroup.structures.parse_atoms import AtomDict, ihex     
from mmgroup.structures.construct_mm import iter_mm       

from mmgroup.structures.ploop import Cocode, PLoop
from mmgroup.structures.autpl import StdAutPlGroup, AutPL ,autpl_from_obj
from mmgroup.structures.construct_mm import iter_mm       
from mmgroup.structures.construct_mm import load_group_name     

try:
    from mmgroup import mat24
except (ImportError, ModuleNotFoundError):
    from mmgroup.dev.mat24.mat24_ref import Mat24
    mat24 = Mat24

from mmgroup.generators import gen_leech2_type
from mmgroup.clifford12 import xsp2co1_elem_to_qs_i, xsp2co1_elem_to_qs 
from mmgroup.clifford12 import xsp2co1_qs_to_elem_i 
from mmgroup.clifford12 import xsp2co1_chain_short_3 
from mmgroup.clifford12 import xsp2co1_neg_elem
from mmgroup.clifford12 import error_string, chk_qstate12
from mmgroup.clifford12 import xsp2co1_unit_elem
from mmgroup.clifford12 import xsp2co1_mul_elem, xsp2co1_inv_elem
from mmgroup.clifford12 import xsp2co1_copy_elem
from mmgroup.clifford12 import xsp2co1_reduce_elem
from mmgroup.clifford12 import xsp2co1_elem_to_leech_op
from mmgroup.clifford12 import xsp2co1_elem_to_bitmatrix
from mmgroup.clifford12 import xsp2co1_set_elem_word 
from mmgroup.clifford12 import xsp2co1_mul_elem_word 
from mmgroup.clifford12 import xsp2co1_xspecial_vector
from mmgroup.clifford12 import xsp2co1_xspecial_conjugate
from mmgroup.clifford12 import xsp2co1_elem_xspecial
from mmgroup.clifford12 import xsp2co1_elem_to_word
from mmgroup.clifford12 import xsp2co1_involution_invariants
from mmgroup.clifford12 import xsp2co1_involution_orthogonal
from mmgroup.clifford12 import xsp2co1_conjugate_elem
from mmgroup.clifford12 import xsp2co1_elem_conjugate_involution
from mmgroup.clifford12 import xsp2co1_elem_subtype

from mmgroup.structures.qs_matrix import QStateMatrix
from mmgroup.structures.construct_mm import iter_strings_from_atoms


FORMAT_REDUCED = True



###########################################################################
# Word class for the group G_{x0}
###########################################################################


MM = None

def import_MM():
    global MM
    if MM is not None:
        return MM
    from mmgroup.mm_group import MM
    return MM


class Xsp2_Co1(AbstractMMGroupWord):
    r"""Models an element the subgroup :math:`G_{x0}` of the monster

    Here  the subgroup :math:`G_{x0}` is a subgroup of the
    monster of structure :math:`2^{1+24}.\mbox{Co}_1`. It is 
    generated by the elements 
    :math:`x_\delta, y_d, y_d, x_\pi, \xi` described in
    section :ref:`mmgroup-label`. 

    The constructor of this class works exactly as the constructor
    of class |MM|. Here all elements of the monster :math:`\mathbb{M}`
    occuring in the constructor must lie in the subgroup :math:`G_{x0}`
    of the monster. So in the constructor all tags are legal, except 
    for the tag ``'t'``. A instance of class  |MM| is accepted in the 
    constructor of this class and vice versa.

    The group operation and the operation on a vector of class
    |MMVector| is the same as in class |MM|.

    This class uses a considerably faster implementation of the
    group operation as in class |MM|, as described in section
    :ref:`group_g_x0_label`.

    The user hardly ever has to deal with this class, since the
    implementation of class |MM| uses the accelerated 
    functions in this class automatically where appropriate.
    """
    MIN_LEN = 16
    __slots__ =  "_data"
    def __init__(self,  tag = None, atom = None, *args, **kwds):
        atoms = iter_mm(self.group, tag, atom, in_G_x0 = True)
        a_atoms = np.fromiter(atoms, dtype = np.uint32) 
        self._data = np.zeros(26, dtype = np.uint64)
        xsp2co1_set_elem_word(self._data, a_atoms, len(a_atoms))
         
    @property
    def data(self):
        return list(map(int, self._data))

    @property
    def short3(self):
        return int(self._data[0])

    @property
    def qs_t(self):
        return QStateMatrix(xsp2co1_elem_to_qs_i(self._data))
        
    @property
    def qs(self):
        return QStateMatrix(xsp2co1_elem_to_qs(self._data))
        
    @property
    def leech_op(self):
        a = np.zeros(576, dtype = np.int8)
        xsp2co1_elem_to_leech_op(self._data, a) 
        return a.reshape((24,24))        

    @property
    def leech_mod2_op(self):
        a = np.zeros(24, dtype = np.uint64)
        xsp2co1_elem_to_bitmatrix(self._data, a) 
        return a        
        
    def order(self, max_order = 119):
        """Return the order of the element of the monster group


        If the argument ``max_order`` is present then the order of the 
        element is checked up to (and including) ``max_order`` only.  
        Then the function returns ``0`` if the order is greater than 
        ``max_order``. By default, the function returns the exact 
        order of the element.
        """
        o = self.qs.order(max_order)
        if o & 1 == 0:
            o = o >> 1
        unit, pw = self.group(), self**o
        if pw == unit:
            return o
        for i in range(2):
            o, pw = 2*o, pw * pw
            if pw == unit:
                return o
        err = "Order of QStateMatrix object not found" 
        raise ValueError(err)


    def as_xsp(self):
        return chk_qstate12(xsp2co1_xspecial_vector(self._data))

    as_Q_x0_atom = as_xsp

    def type_Q_x0(self):
        r"""Return type of element if it is in the subgroup :math:`Q_{x0}`

        If the element is in the subgroup :math:`Q_{x0}` of the monster
        then the function returns the type of the vector in the Leech 
        lattice modulo 2 corresponding to this element. That type is
        0, 2, 3, or 4.

        The function raises ValueError if the element is not
        in the subgroup :math:`Q_{x0}`. 
        """
        v = self.as_Q_x0_atom()
        return gen_leech2_type(v)

    def xsp_conjugate(self, v, sign = True):
        v = np.array(v, dtype = np.uint64, copy=True)
        shape = v.shape
        assert len(shape) <= 1
        v = np.ravel(v)
        chk_qstate12(
            xsp2co1_xspecial_conjugate(self._data, len(v), v, sign))
        if len(shape):
            return list(map(int,v))
        else:
            return int(v[0])

    def mul_data(self, data):
        a_atoms = np.array(data, dtype = np.uint32)
        xsp2co1_mul_elem_word(self._data, a_atoms, len(a_atoms))
        return self


    @property
    def mmdata(self):
        a = np.zeros(10, dtype = np.uint32)
        length = chk_qstate12(xsp2co1_elem_to_word(self._data, a))
        return a[:length]

    def _involution_invariants(self):
        """Wrapper for C function  xsp2co1_involution_invariants"""
        invar = np.zeros(12, dtype = np.uint64)
        xsp2co1_involution_invariants(self._data, invar)
        v1 = xsp2co1_involution_orthogonal(invar, 1)
        v0 = xsp2co1_involution_orthogonal(invar, 0)
        return invar, v1, v0


    def conjugate_involution(self, mmgroup = None):
        r"""Find an element conjugating an involution standard element

        If the element :math:`g` given by ``self`` is an involution 
        in  the monster then the method computes an element :math:`h` 
        of the monster   with  :math:`h^{-1} g h = z`, where  
        :math:`z` is define as follows:

        If :math:`g = 1`, we put :math:`h = z = 1`

        if :math:`g` is a 2A involution (in the monster) then we let
        :math:`z` be the involution in  :math:`Q_{x0}` corresponding 
        to the Golay cocode word with entries  :math:`2,3` being set.

        if :math:`g` is a 2B involution (in the monster) then we let 
        :math:`z` be the central involution in :math:`G_{x0}`

        The function returns a pair ``(I, h)``, where :math:`h` as an 
        element of the instance  ``MM`` of class ``MMGroup``. We put
        ``I = 0`` if :math:`g = 1`. We put ``I = 1, 2`` if 
        :math:`g` is a 2A or 2B involution, respectively.

        The function raises ``ValueError`` if :math:`g` is not an 
        involution. 

        This is a wrapper for the C 
        function ``xsp2co1_elem_conjugate_involution``.

        Parameter ``mmgroup`` is not for public use. In special cases
        it may specify an alternative class implementing the monster 
        group. Then the function returns :math:`h` as an instance of 
        that class.
        """
        if mmgroup is None:
             mmgroup = MM if MM is not None else import_MM()            
        a = np.zeros(15, dtype = np.uint32)
        len_a = xsp2co1_elem_conjugate_involution(self._data, a)
        chk_qstate12(len_a)
        return (len_a >> 8), mmgroup('a', a[:len_a & 0xff])



    def _qs_v2(self):
        i = 26
        while i and int(self._data[i-1] == 0):
            i -= 1
        i -= 14
        assert i & 1 == 0 and 0 <= i <= 12
        return i >> 1
        
    @property
    def subtype(self):
        r"""Return subtype of an element

        Let :math:`g \in G_{x0}` be stored in ``self``. The function
        returns the subtype of a :math:`g`. If :math:`g` maps the 
        standard frame :math:`\Omega` of the Leech lattice modulo 2 
        to a frame of subtype :math:`t` then :math:`g` has 
        subtype :math:`t`.

        The subtype is returned as a pair of integers as in the 
        corresponding method in class |XLeech2|, see section 
        :ref:`computation-leech2` in the **guide** for background.

        Since the subtype is determined by the size of the denominators
        of the representation :math:`4096_x`, it can be computed very 
        fast.
        """
        res = xsp2co1_elem_subtype(self._data)
        assert res & 0x40 == 0x40
        return res >> 4, res & 0xf

    def str(self):
        """Convert group element to a string
        """
        atoms = self.mmdata
        strings = iter_strings_from_atoms(atoms, abort_if_error=0)
        s = "*".join(strings) 
        return "Xsp2_Co1<%s>" % (s if len(s) else "1")



###########################################################################
# The class representing the group G_x0
###########################################################################



def cocode_to_xsp2co1(g, c):
    res =  g.word_type(group = g)
    chk_qstate12(xsp2co1_elem_xspecial(res._data, c.value & 0xfff))
    return res

def ploop_to_xsp2co1(g, pl):
    res =  g.word_type(group = g)
    value = (c.value & 0x1fff)
    value = (value << 12) ^ mat24.ploop_theta(value)
    chk_qstate12(xsp2co1_elem_xspecial(res._data, value))
    return res


def autpl_to_xsp2co1(g, aut):
    res =  g.word_type(group = g)
    a = np.zeros(2, dtype = uint32)
    a[0] = 0x10000000 + (aut._cocode & 0xfff)   # tag 'd'
    a[1] = 0x20000000 + aut._perm_num           # tag 'p'
    chk_qstate12(xsp2co1_set_elem_word(res._data, a, 2))
    return res

def mmgroup_to_xsp2co1(g, mm):
    res =  g.word_type(group = g)
    chk_qstate12(xsp2co1_set_elem_word(res._data, mm._data, mm.length))
    return res


def xsp2co1_to_mm(mmgroup, xsp):
    g = mmgroup()
    g._extend(10)
    g.length = chk_qstate12(xsp2co1_elem_to_word(xsp._data, g._data))
    g.reduced = 0 
    return g



@singleton
class Xsp2_Co1_Group(AbstractMMGroup):
    r"""Model the subgroup :math:`G_{x0}` of the Monster
    
    The group :math:`G_{x0}` is the subgroup of the Monster group
    genertated by the elements with tags ``p, d, x, y, l`` in the  
    class |MMGroup| representing the Monster group.  :math:`G_{x0}`
    has structure :math:`2**(1+24).\mbox{Co}_1` in ATLAS
    notation, see :cite:`Con85`, :cite:`Atlas`.
       

    
    :param \*data:

      A variable number of arguments; each argument describes an
      element of the subgroup ``2**(1+24).Co_1`` of the monster.
      These elements are multiplied.  
        

    Depending on its type each parameter in **\*data** is  
    interpreted as follows:

    .. table:: Legal types for constructor of class ``Xsp2_Co1_Group``
      :widths: 25 75

      ===================== ==================================================
      type                  Evaluates to
      ===================== ==================================================
      tuple (``tag, data``) Create an element of the monster group as 
                            described in class |MMGroup|. Here legal tags
                            are: ``'p', 'd', 'x', 'y', 'z', 'l'``.  The
                            resulting element must lie in the subgroup
                            :math:`G_{x0}` of the monster.
                            
      class |MMGroup|       Create an element of the Monster group. That                
                            element must lie in the subgroup
                            :math:`G_{x0}` of the monster.

      class |AutPL|         Create an element of the subgroup of
                            :math:`G_{x0}` given by class |AutPL|.
                           
      class ``Xsp2_Co1``    Create a copy of an element of this class.
                           
      pair (``qstate, x``)  Deprecated and not implemented!!!!!
                            This kind of construction is for testing and
                            not recommended for public use. Here        
                            ``qstate`` must be an instance of class 
                            ``mmgroup.structures.qs_matrix.QStateMatrix``
                            and ``x`` must be an integer representing a
                            short Leech lattice vector modulo 3.
      ===================== ==================================================

        
    :raise:
        * TypeError if ``type(data)`` is not as expected.
        * ValueError if ``data`` cannot be converted to an
          instance of class  `` Xsp2_Co1``.
    
    
    """
    __instance = None
    __slots__ = "data"
    STD_V3  = 0x8000004
    word_type = Xsp2_Co1
    is_mmgroup = True


    def __init__(self):
        """ TODO: Yet to be documented     


        """
        super(Xsp2_Co1_Group, self).__init__()
        self.atom_parser = AtomDict(self.atom)


    def atom(self, tag = None, i = "r"):
        return self.word_type(tag, i)

    def _imul(self, g1, g2):
        chk_qstate12(xsp2co1_mul_elem(g1._data, g2._data, g1._data))
        return g1

    def _invert(self, g1):
        w = self.word_type()
        chk_qstate12(xsp2co1_inv_elem(g1._data, w._data))
        return w

    def copy_word(self, g1):
        w = self.word_type()
        xsp2co1_copy_elem(g1._data, w._data)
        return w

    def reduce(self, g1, copy = False):
        if copy:
            g1 = self.copy_word(g1)
        chk_qstate12(xsp2co1_reduce_elem(g1._data))
        return self
       
    def _equal_words(self, g1, g2):
        chk_qstate12(xsp2co1_reduce_elem(g1._data))
        chk_qstate12(xsp2co1_reduce_elem(g2._data))
        return (g1._data == g2._data).all()

    def _embed_number(self, n):
        w = self.word_type()
        if (n == -1):
            xsp2co1_neg_elem(w._data)
            n = 1
        if n == 1:
            return w
        raise TypeError("Cannot convert a number to a group element")
        
    def from_qs(self, qs, x):  
        w = self.word_type()
        w0 =  xsp2co1_qs_to_elem_i (qs, x)
        for i in range(26):
             w._data[i] =  w0[i]
        return w             

    def str_word(self, v1):
        #return "Xsp2_Co1 " + str_xsp2_co1(v1._data)
        atoms = v1.mmdata
        strings = iter_strings_from_atoms(atoms, abort_if_error=0)
        s = "*".join(strings) 
        return "Xsp2_Co1<%s>" % (s if len(s) else "1")
 
    def from_xsp(self, x):
        w = self.word_type()
        chk_qstate12(xsp2co1_elem_xspecial(w._data, x))
        return w

    def from_data(self, data):
        """Create a group element from an array of generators

        Internally, an element of group is represented
        as an array of unsigned 32-bit integers, where each entry
        of the array describes a generator. See section
        :ref:`header-mmgroup-generators-label` for details.
 
        This function creates an element of the group from
        such an array of integers.

        :param data: An array-like object representing a 
                     word of generators of the monster group

        :return: An element of this instance of the group 
        :rtype:  an instance of class 
                 mmgroup.structures.xsp2_co1.Xsp2_Co1_Word

        """
        return self.word_type('a', data)




_dict_pm3 = {0: '0', 1:'+', 0x1000000:'-', 0x1000001:'0'}
def str_leech3(x):
    x = int(x)
    lst = [_dict_pm3[(x >> i) & 0x1000001] for i in range(24)]
    return "(" + "".join(lst) + ")"

def str_xsp2_co1(data, factor = 1, t = False):
    qs0 = xsp2co1_elem_to_qs_i(data) if t else xsp2co1_elem_to_qs(data)
    qs = QStateMatrix(qs0) / factor
    return str_leech3(data[0]) + " (x) " + str(qs)


try:
    from mmgroup.clifford12 import xsp2co1_error_pool
    def get_error_pool(length):
        assert length > 0
        a = np.zeros(length, dtype = np.uint64)
        length = xsp2co1_error_pool(a, length)
        return list(map(int, a[:length])) 
except:
    def get_error_pool(length):        
        return []    





StdXsp2_Co1_Group = Xsp2_Co1_Group()
Xsp2_Co1.group = StdXsp2_Co1_Group
load_group_name(StdXsp2_Co1_Group, "MX")



   


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import numpy as np
from random import randint

import pytest

from mmgroup.mm import mm_sub_test_prep_pi_64
from mmgroup import mat24





def get_prep_pi64(delta, pi):
    """Obtain component tbl_perm64 from function mm_sub_prep_pi()

    More specifically, function mm_sub_prep_pi() in module 
    mm_tables.c takes a pair (delta pi) indicating a automorphism
    of the Parker loop. That function computes a structure
    containing a component 'tbl_perm64'. This function performs
    the corresponding computation for the input (delta, pi) and
    return component 'tbl_perm64' of the computed structure
    as a numpy array of length 2*759 with dtype = uint32.

    For the interpretation of the returned data, see documentation 
    of function mm_tables.c
    """
    a = np.zeros(759*7, dtype = np.uint32)
    mm_sub_test_prep_pi_64(delta, pi, a)
    return a.reshape(759,7)

def parse_a(a):
    """Interpret data returned by function get_prep_pi64()

    These data describe an automorpism of the Parler loop.

    The function returns a list of tuples (source, sign, fields),
    one for each octde 0,...,759-1. 'source' is the number of
    the octad which is the preimage of octad i. 'sign' is the
    sign of the image of octad 'source'. 'fields' is a 6-tuple
    of 6-bit integers describing the mapping of the suboctads,
    see documentaiton in file mm_tables.c.
    """
    a1 = [None] * 759
    for i, row in enumerate(a):
        idx = row[0] & 0x3ff
        sgn = (row[0] >> 12) & 1
        a1[i] = (idx, sgn, [int(i) for i in row[1:]])
    return a1


def dump_pi64(delta, pi, a, format = 0):
    print("raw data for delta = 0x%x, pi = %d:" % (delta, pi))
    for i in range(759):
         print("%3d: %s" % (i, a[i]))



def map_octad(oct, delta, pi):
    """Map octad through a Parker loop automorphsim to octad.

    Here 'oct' is an octad given in ocatad representation,
    interpreted as a (positive) element of the Parker loop.
    The pair (delta, pi), with delte a cocode element and pi
    the number of a permutation in Mat24, is an automorphism
    of the Parker loop. 

    The function returns a pair (sign, img_octad) representing
    an element of the Parker loop with a given sign. img_octad 
    is the number of the correspondig octad in octad 
    representation.

    The return value is the image of the octad under the
    Parker loop automorphism (delta, pi).
    """
    gcode = mat24.octad_to_gcode(oct)
    perm = mat24.m24num_to_perm(pi)
    m = mat24.perm_to_autpl(delta, perm)
    image = mat24.op_ploop_autpl(gcode, m)
    image_oct = mat24.gcode_to_octad(image)
    sign = (image >> 12) & 1
    return sign, image_oct
 
def suboctad_to_cocode(suboctad, oct):
    """Compute code vector from octad and suboctad

    The octad 'oct' must be given in 'octad' representation and
    the suboctad must be given as in function 
    mat24_suboctad_to_cocode in module mat24_functions.c.
    The function returns the suboctad as a cocode element in 
    'cocode' representation.    
    """
    gcode = mat24.octad_to_gcode(oct)
    return  mat24.suboctad_to_cocode(suboctad, gcode)
   

def map_cocode(c, pi):
    """Map cocode element with permutation pi

    'c' is a cocode element in 'cocode' representation.
    'pi' is the number of a permutation in the Mathieu group
    Mat24. The function returns the image of 'c' under the 
    permutation 'pi' in 'cocode' representation.
    """
    perm = mat24.m24num_to_perm(pi)
    return mat24.op_cocode_perm(c, perm)


def one_test_prep_pi(delta, pi, verbose = 0):
    a = get_prep_pi64(delta, pi)
    a1 = parse_a(a)
    if verbose:
        print("testing delta = %x, pi = %d" % (delta, pi))
        if verbose > 2:
            dump_pi64(delta, pi, a)
        if verbose > 1:
            for i, l in enumerate(a1): print(i, l)
    for i in range(100):
        oct = randint(0,758)
        preimage = a1[oct]
        pre_oct, pre_sign, pre_suboctads = preimage
        sign, image_oct =  map_octad(pre_oct, delta, pi)
        assert image_oct == oct, (image_oct, oct)
        for j, pre_suboctad in enumerate(pre_suboctads):
            suboctad = 2*2**j - 1
            pre_cocode = suboctad_to_cocode(pre_suboctad, pre_oct)
            cocode = suboctad_to_cocode(suboctad, oct)
            img_cocode = map_cocode(pre_cocode, pi)
            #print(j, pre_suboctad, suboctad, pre_cocode, cocode, pre_oct, oct)
            assert cocode == img_cocode, (i, j, cocode, img_cocode)
        assert sign == pre_sign, (sign, pre_sign)

def pi_testcases():
    testdata = [
        (0, 0),
    ] 
    for x in testdata: yield x  
    for i in range(80):
        yield randint(0, 0xfff), randint(0, mat24.MAT24_ORDER - 1)


def even_testcases():
    testdata = [
        (0, 0),
    ] 
    for x in testdata: yield x  
    for i in range(80):
        yield randint(0, 0xfff) & 0x7ff, randint(0, mat24.MAT24_ORDER - 1)


@pytest.mark.mm_op
@pytest.mark.parametrize("testcases", [pi_testcases])
def test_prep_pi(testcases, verbose = 0):
    e = "\n" if verbose else ""
    print("Test function mm_sub_prep_pi... ", end = e)
    for delta, pi in testcases():
        one_test_prep_pi(delta, pi, verbose = verbose) 
    print("passed")



    


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import numpy as np
from numbers import Integral
from random import randint


import pytest


from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepSpace
from mmgroup.structures.abstract_mm_rep_space import sparse_subspace
from mmgroup.mm_space import characteristics
from mmgroup.mm import mm_aux_index_sparse_to_extern
from mmgroup.mm import mm_aux_index_extern_to_sparse
from mmgroup.mm import mm_aux_mmv_extract_sparse

from mmgroup.tests.spaces.spaces import MMTestSpace
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.groups.mgroup_n import MGroupNWord


#from mm_vector import FastMmRepSpace, PRIMES
#from mm_vector import SparseMmRepVector
#from mm_vector import AbstractRepVector
#from mm_vector import mm_sparse_purge
#from mm_vector import basis_vectors_to_sparse

PRIMES = characteristics()


np.set_printoptions(formatter={'int':hex})

########################################################################
# group_blocks
########################################################################

"""Finding blocks in the matrix of an element g of the monster

Let R be the representation of the monster Mm and let g in Mm. We want 
to test the fast C implementation of the operation of g against the 
slow python implementation. While the C implementation always operates 
on a vector v in R of full length, the python implementation works on 
the individual components ov v, with run time depending on the number 
of componennts of v. So we select a uniform random vector v and an 
atom of the group g and we let the C implementation of g operate of v.
Then we let the python implementation of g operate on a few components
of v selected at random, and we compare the results. Of course, one
component of g(v) may depend on several components of v, and we
must apply the operation of g to (at least) all these components
for obtaning a correct component of g(v).

Let B be a (usually small) set of basis vectors of R and let g be 
an element of Mm. Then function group_blocks() maps B to a superset 
B_g of B of basis vectors of R with the following property:

g maps the linear span of B_g to a subspace spanned by basis vectors.

For a random vector v in R, let v1 be the the projection of v onto
the linear span of B_g. Then g(v1) is the projection of g(v) onto the 
space spanned by the basis vectors for which the coordinate in 
g(v1) is not zero.

If B_g is not too large, then g(v1) can also be computed with the 
python implementation without too much effort. This makes it
possible to test C implementation against the python implementation.

In the worst case, B_g may be the whole basis of R. Function 
group_blocks() raises ValueError if the computation of B_g is too
complicated. The function returns with success if the following 
conditions are met:

  -  g contains at most one non-monomial atom. Non-monomial 
     atoms are those with tag 't' or 'l'.
  -  An atom with tag 'l' has no direct or indirect predecessor 
     with tag 'p'.

"""


def group_blocks_t(a):
    """Auxiliary function for function group_blocks()

    Assuming that the numpy array 'a' aencodes a basis B, the 
    function returns the expanded basis B_g in case that an  
    atom with tag 't' has been found in the group element.

    The function is implemented as a generator. It yields
    unsigned 32-bit integers representing basis vectors.
    """
    ABC_tags = set()
    T_tags = set()
    for sp in a:
        tag = (sp >> 25) 
        if 1 <= tag <= 3:
            i, j = (sp >> 14) & 0x7ff, (sp >> 8) & 0x3f
            if i != j:
                ABC_tags.add(sp & 0x1ffff00)
            else:
                yield sp
        elif tag == 4:
            T_tags.add(sp & 0xffffc000 )
        else:
            yield sp
    for sp in ABC_tags:
        for tag in (1,2,3):
            yield (tag << 25) + sp
    for sp in T_tags:
        for j in range(64):
            yield sp + (j << 8)
            

def group_blocks_l(a):
    """Auxiliary function for function group_blocks()

    Assuming that the numpy array 'a' ancodes a basis B, the 
    function returns the expanded basis B_g in case that an  
    atom with tag 'l' has been found in the group element.

    The function is implemented as a generator. It yields
    unsigned 32-bit integers representing basis vectors.
    """
    A_tags = set()
    ZY_tags = set()
    for sp in a:
        tag = (sp >> 25) 
        if tag == 1:
            A_tags.add(sp & 0xffff1c00)
        elif tag in (6,7):
            ZY_tags.add(sp & 0xfffc1c00)
        else:
            yield sp
    for sp in A_tags:
        i, j = (sp >> 14) & 0x7ff, (sp >> 8) & 0x3f
        if i != j:
            for i0 in range(4):
                for j0 in range(4):
                    yield sp + (i0 << 14) + (j0 << 8)
        else:
            for i0 in range(4):
                for j0 in range(i0 + 1):
                    yield sp + (i0 << 14) + (j0 << 8)
    for sp in ZY_tags:
        for i0 in range(16):
            for j0 in range(4):
                yield sp + (i0 << 14) + (j0 << 8)




def group_blocks(basis_vectors, g):
    """Find block in matrix g operating on the basis_vectors.

    Here g in a group element of the monster group Mm, where the
    monster group is an instance of class MgroupN.

    'basis_vectors' represents a set B of basis vectors of the
    representatiom of Mm. It may be

     - a vector in a representation R_p of Mm, where R_p is an
       instance of (a subclass of) class AbstractMmRepSpace. 
       Then all basis vectors corresponding to nonzero components 
       of that vector are taken
     - A string of tags. Then for each tag a random basis
       vector with that tag is generated.
     - An array like object that will be converted to a one-
       dimensional numpy array of dtype np.uint32. This array
       is interpreted as a sparse representation of a vector v.
       In this case all basis vectors contained in the sparse 
       representation ar taken, ignoring coordinates.


    Given the set B of basis vectors and an element g of Mm, the 
    function returns a list B_g of basis vectors in sparse format
    as a numpy array of dtype = np.uin32.

    Here B_g is a superset the set B of basis vectors with the 
    following property:

    g maps the linear span of B_g to a subspace of the 
    representation of Mm spanned by basis vectors.

    The function may raise ValueError if the computation of B_g 
    is too complicated, see comment in the header of this 
    section for details. 
    """
    if isinstance(basis_vectors, str):
        a = sparse_subspace(*((x,) for x in basis_vectors))
    else:
        a = sparse_subspace(basis_vectors)
    # Now a is the list of basis vectors encoded as a numpy array.
            
    dirty_t = dirty_l = error = False
    # Here dirty_t (or dirty_l) means that we can no longer control
    # the proliferation of components due to coming atoms with 
    # tag 't' (or tag 'l').
    expansion_function = None

    for tag, i0 in g.as_tuples():
        if tag == 'p' and i0:
            dirty_l = True 
        elif  tag == 't' and i0 % 3:
            dirty_l = True 
            error = error or dirty_t
            expansion_function = group_blocks_t
        elif  tag == 'l' and i0 % 3: 
            dirty_t = True 
            error = error or dirty_l
            expansion_function = group_blocks_l
    if error:
        raise ValueError("Matrix too complicated for finding blocks")
    if expansion_function:
        a = np.fromiter(expansion_function(a), dtype = np.uint32)
    return a
        


########################################################################
# tests with a sparse vector v
########################################################################


def one_test_op(v, g, f_mul = None, verbose = 0):
    """ Test operation of group element g on unit vector v

    Here v is a vector in a space, such that v is an 
    instance of class SparseMmVector. g is an element of the 
    group v.space.group operating on the space v.space. 

    The group operation v -> v * g is tested against the operation
    sparse_space(v) * g, where sparse_space(v) is an instance of
    class SparseMmRepVector, which represents the same space as
    v.space, but with vectors given in sparse form.

    If f_mul is set to a function such that f_mul(v, g) = v * g
    holds, an alternative implemetation of the group multiplication
    may be tested. This is useful for some low-level module tests. 
    """
    #print("test_op_case")
    space = MMTestSpace(v.p)
    ref_space = space.ref_space
    v_sp =  ref_space(v)
    if verbose:
        print("p= %d, v=%s, g=%s" % (v.p, v, g))
        print("v_sparse=", v_sp)
    vg = f_mul(v, g) if f_mul else v * g 
    vg.check()
    vg_sp = ref_space(vg) 
    v_sp_g = v_sp * g
    ok = v_sp_g == vg_sp
    if verbose or not ok:
        if not ok:   
            print("p= %d, v=%s, g=%s" % (v.p, v, g))
        if 1:
            print("v * g  expected in sparse format:")
            print(v_sp_g)
            print("v * g  obtained in sparse format:")
            print(vg_sp)            
        print("v*g=", vg) 
        if not ok:   
            raise ValueError("Group operation on vector failed")
    vg_converted = space(vg_sp)
    if vg != vg_converted:
        raise ValueError("Conversion of result from sparse format failed")
 
 

def op_testcases(p):
    space = MMTestSpace(p)
    group = space.group
    for v_tag in "XYZ":  
        for g_tag in "d": 
            for i0 in (1 <<  i for i in range(11)):  
                for i in range(10):
                    g = group(g_tag, 'r')
                    v = space(v_tag, i0, randint(0,23))
                    yield v, g
    for v_tag in "ABCXYZT":  
        for g_tag in "dpxy":   
            for i in range(10):
                v = space(v_tag, 'r')
                g = group(g_tag, 'r')
                yield v, g
        for g_tag in "tl":   
            for e in [1,2,1,2]:
                v = space(v_tag, 'r')
                g = group(g_tag, e)
                yield v, g
 
   
@pytest.mark.mm_op
def test_op(f_mul = None, verbose = 0):
    print("Testing group operation on sparse vectors")
    i = 0
    for p in PRIMES:
        for args in op_testcases(p):
            if verbose: print("Test", i+1)
            one_test_op(*args, f_mul = f_mul, verbose = verbose)
            i += 1
            #print("<%d>" % i, end = "", flush=True)
    print("Test passed")
   
   

########################################################################
# tests with a random vector v
########################################################################


def one_test_rand_op(v, g, basis_vectors, f_mul = None, verbose = 0):
    """Test operation of group element g on random vector v

    We compute v * g with the fast C and with the slow python 
    implementation and check that the results are equal.

    v must be a vector in a space which is an instance of class
    MMTestSpace, and g must be an element of the monster group
    such that v * g can be computed. So g may e.g. be an element
    of v.space.group. If v is an instance of class MMTestSpace
    then a random vector in the space given by v is created.

    Since the python implementation is rather slow, we compute v*g
    in python only for the projection of v to the space spanned 
    by the given 'basis_vectors'. Here a set of basis vectors
    must be encoded in the same way as in function group_blocks().

    If g operates non-monomially on v, it may be necessary to
    consider more basis vectors than given by the argument
    'basis_vectors'. Then function group_blocks() is used to
    extend the set of basis vectors in suitable way if necesary.

    The extension of basis vectors may fail if the structure of
    the matrix of g is too complicated. But that extension is
    always successful if g contains at most one atom or if all
    atoms in g operate monomially. So we can test the operation
    of all atoms in g.

    If f_mul is set to a function such that f_mul(v, g) = v * g
    holds, an alternative implemetation of the group multiplication
    may be tested in te same way as in function test_op_case(). 
    """
    #if isinstance(v, AbstractMmRepSpace):
    #    v = v.rand_vector()
    space = MMTestSpace(v.p)
    ref_space = space.ref_space
    v_g = f_mul(v, g) if f_mul else v * g
    if verbose:
        print("p = %d, g = %s" % (v.p, g))
    basis_block = group_blocks(basis_vectors, g)
    v_sparse = ref_space('S', v.get_sparse(basis_block))
    if verbose:
        print("\nTest operation v * g")
        print("len(v_sparse) =", len(v_sparse))
        print("sparse(v) =", v_sparse)
    v_sparse_g = v_sparse * g
    v_g_sparse = ref_space(v_g.projection(*v_sparse_g.as_tuples()))
    #v_g_sparse = ref_space(v_g.projection(v_sparse_g))
    ok  = v_sparse_g == v_g_sparse
    if verbose or not ok:
        if not ok:
            print("p = %d, g = %s" % (v.p, g))
            print("basis block = ", v.get_sparse(basis_block))
            print("len(sparse v) =", len(v_sparse),
                            ", type =", type(v_sparse))
            print("sparse(v) =", v_sparse)
            print("len(sparse(v*g)) =", len(v_g_sparse))
            print("sparse(v * g) =", v_g_sparse)
        print("len(sparse(v) * g) =", len(v_sparse_g))
        print("sparse(v) * g =", v_sparse_g)
        print("diff =", v_g_sparse - v_sparse_g)
        print("")
    if not ok:
        raise ValueError("Group operation in C differs from python op.")
    
    

@pytest.mark.mm_op
def test_rand_op(n_tests = 3, f_mul = None, verbose = 0):
    print("Testing group operation on random vectors")
    for i in range(n_tests):
        for p in PRIMES:
            space = MMTestSpace(p)
            group = space.group
            for atom in "dpxytl":
                g = group(atom, "n")
                if atom == "l":
                    basis = "D" + "BCTX" * 20 + "A" * 4 + "YZ" * 2
                elif atom == "t":
                    basis = "D" * 3 + "ABC" * 7 +  "T" * 2 + "XYZ" * 20
                else:
                    basis = "D" * 3 + "ABC" * 10 + "TXYZ" * 20 
                for i in range(n_tests): 
                    v = space("R")               
                    one_test_rand_op(v, g, basis, f_mul = f_mul,
                                  verbose = verbose)
    print("Test passed")


########################################################################
# tests operation on tag 'A' of the vector
########################################################################


def tag_A_testwords():
    for atom in "dpxyl":
        yield [(atom, "n")]
    yield [(c, "n") for c in "dpxyl" * 3]


@pytest.mark.mm_op
def test_rand_op_tag_A(n_tests = 4, f_mul = None, verbose = 1):
    print("Testing group operation on random vectors")
    for i in range(n_tests):
        for p in PRIMES:
            space = MMTestSpace(p)
            group = space.group
            for w in tag_A_testwords():
                g = group(w)
                #print(g)
                v = space('R')
                a_g = v.copy()
                v_g = v * g
                len_g = len(g.mmdata)
                res = v.ops.op_word_tag_A(a_g.data, g.mmdata, len_g, 1)
                assert res == 0
                #print(v_g['A'] - a_g['A'] + 3)
                assert (v_g['A'] ==  a_g['A']).all()
                res = v.ops.op_word_tag_A(a_g.data, g.mmdata, len_g, 2)
                assert res == 0
                res = v.ops.op_word_tag_A(a_g.data, g.mmdata, len_g, -3)
                assert res == 0
                assert (v['A'] ==  a_g['A']).all()
    print("Test passed")








from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.structures.qs_matrix import QStateMatrix
from mmgroup.structures.qs_matrix import qs_unit_matrix, qs_rand_matrix
from mmgroup.structures.qs_matrix import qs_column_monomial_matrix
from mmgroup.structures.qs_matrix import qs_row_monomial_matrix
from mmgroup.structures.qs_matrix import FORMAT_REDUCED

#####################################################################
# Convert matrix of type QStateMatrix to a complex matrix
#####################################################################

def eval_A_Q(ncols, data, v):
    r"""Auxiliary function for function ``slow_complex``.
    
    Evaluate parts ``A``and ``Q`` of a matrix for a vector ``v``.
    
    Given a state matrix ``qs`` type ``QStateMatrix``, parameter
    ``ncols`` should be equal to ``qs.rows + qs.cols``, and parameter
    ``data`` should be equal to ``qs.data``.  Then parts ``A``and ``Q``
    of ``qs`` are evaluated for a bit vector ``v`` of bit length
    ``qs.nrows``, where ``qs.nrows`` is equal to ``len(data)``.
    
    ``A`` is the ``qs.nrows`` times ``qs.ncols`` bit matrix that
    consistts of the first `qs.ncols`` columns of bit matrix ``data``.
    ``Q`` is the ``qs.nrows`` times ``qs.nrows`` bit matrix that
    consistts of the following `qs.rows`` columns of matrix ``data``.
    
    The function returns a pair ``(a, q)``, where ``a`` is the matrix
    product :math:``v \cdot A``.  ``q`` is defined as follows:
    Let ``R`` be the bit matrix of the same shape as ``Q``
    with ``R[i,j] = v[i] * v[j] * Q[i,j]``. Then ``q`` is the
    sum of all bits in ``R`` modulo 4
    
    We pass the entries ``ncols,  data`` of the state matrix
    ``qs`` instead of ``qs`` itself, for reasons of efficiency.
    """
    # Clear irrelevant bits of v 
    v &= (1 << len(data)) - 1  
    # Sum up  :math:``v \cdot A`` in ``a``, the off-diagonal elements 
    # of ``R`` in ``q``, and the diagonal elements of ``R`` in ``diag``. 
    # Since ``R`` is symmetric, we XOR  the lower triagular off-diagonal 
    # elements of ``R`` in ``q`` and double the result.
    a = q = diag = 0
    for i, d in enumerate(data):
        # ``d`` = row ``i`` of ``(A,Q)`` if bit ``i`` of ``v``
        # is set an ``d = 0`` otherwise.        
        d &=  -((v >> i) & 1)
        # XOR  (the ``A`` part of)  ``d`` to ``A``.
        a ^= d
        # Next let ``d`` be row ``i`` of the ``R``.
        d = (d >> ncols) & v
        # Add diagonal bit of row ``i`` of ``R`` to ``diag``
        diag += (d >> i) & 1
        # XOR lower triangular part of that row to ``q``
        q ^= d &  ((1 << i) - 1)
    # Calculate the bit parity of ``q`` in bit 0 of ``q``
    for sh in [32, 16, 8, 4, 2, 1]:
        q ^= q >> sh 
    # Mask out the relevant bits of the result ``a``
    a &= (1 << ncols) - 1
    # Caclulate ``q`` as the doubled sum of the bits the lower 
    # triangular part of the bits of ``R``, plus the sum
    # of the diagonal bits of ``R``. 
    q = (2*q + diag) & 3
    return a, q   




def slow_complex(m):
    """Convert matrix ``m`` of type ``QStateMatrix`` to complex
    
    The function returns a complex ``numpy`` array of appropriate
    shape containing the matrix represented by the input ``m``.
    
    This is a pathon implementation for testing the fast
    method ``m.complex()``    
    """
    # Compute the complex scalar factor ``f``  from ``m.factor``.
    e, phi = m.factor 
    f = 2.0**(0.5*(e - phi)) * (1+1j)**phi
    # Put ``phases[i] = sqrt(-1)**i * f,  0 <= i < 4.``
    phases = [f * 1j**i for i in range(4)]
    # Create comples zero array ``a`` od appropriate length.
    a = np.zeros(1 << m.ncols, dtype = complex)
    # For all vectors of length ``m.nrows`` with bit 0
    # of ``v`` equal to 1 put ``index =  v * A``, and  let ``q``
    # be the sum of the matrix ``R`` derivend from ``v`` and
    # ``Q`` as in function ``eval_A_Q()``.
    # Then add ``phases[q mod 4]`` to entry ``a[index]``
    ncols, data = m.ncols, m.copy().check().raw_data[:m.nrows]
    for v in range(1, 1 << len(data), 2):
        index, q = eval_A_Q(ncols, data, v)
        a[index] += phases[q]
    # Change shape of array ``a`` as required
    rows, cols = m.shape
    return a.reshape((1 << rows, 1 << cols))



#####################################################################
# Check echelon form  matrix of type QStateMatrix
#####################################################################


def check_echelon(m, reduced = False):
    """Check if a matrix ``m`` is in echelon form

    ``m`` must be a matrix of type ``QStateMatrix``.
    If ``reduced`` is True them the matrix is checked to be
    in reduced echelon form.

    Raise ``ValueError`` if matrix ``m`` is not in correct
    echelon form.
    """
    err = "Bad vector in row % of part A of QSMatrix"
    data = m.copy().check().raw_data[:m.nrows]
    mask = (1 << m.ncols) - 1
    index = m.ncols + 1
    for i, v in enumerate(data[1:]):
        v &= mask
        new_index = v.bit_length()
        if new_index >= index:
            print("\nError: " + err % i)
            print(m)
            err1 = "QSMatrix is not in echelon form"
            raise ValueError(err1)
        index = new_index
        if reduced:
            v = reduce(__or__, data[:i], 0) 
            if (v & (1 << (index - 1))):
                print("\nError: " + err % i)
                print(m)
                err1 = "QSMatrix is not in reduced echelon form"
                raise ValueError(err1)
                
            
        
        


#####################################################################
# Check that states are approcimately equal
#####################################################################

EPS = 1.0e-8

def check_eq_qstate(m1, m2, text = None):
    """Raise ValueError unless matrices ``m1, m2``  are equal.
    
    Inputs ``m1, m2`` must be matrices of type ``QStateMatrix``.
    Both matrices are converted to complex matrices (with a
    method of class ``QStateMatrix`` that does not reduce the
    matrices) and testes for (approximate) equality.   

    ``text`` is an optional text to be displayed in case of
    an error.    
    """
    c1 = m1.copy().complex()
    c2 = m2.copy().complex()
    eq = np.amax(abs(c1 - c2), initial = 0.0) < EPS
    if not eq:
        err = "Error in comparing instances of QStateMatrix"
        print("\n" + (text if text else err) + "\n")
        print("m1 = ", m1)
        print("m2 = ", m2)
        if max(m1.shape + m2.shape) < 4:
            print("complex m1 = \n", c1)
            print("complex m2 = \n", c2)
        raise ValueError(err)
         

def compare_complex(c1, c2, text, qstate = None):
    """Check that two complex matrices are equal
    
    ``c1`` is the expected and ``c2`` is the obtained complex matrix.
    ``qstate`` is an optiona state representing the matrix.
    """
    try:
        diff = np.amax(abs(c1 - c2), initial = 0.0)
    except:
        err = "\nCannot compare arrays of shapes %s and %s\n"
        print(err % (c1.shape, c2.shape))
        raise
    if diff > EPS:
        err = text if text else "Error in comparing complex matrices"
        print("\n" + err + "\n")
        print("qstate  = ", qstate)
        if max(sum(c1.shape), sum(c2.shape)) <= 64:
            print("expected\n", c1)
            print("obtained\n\n", c2)
            print("expected - obtained\n\n", c1 - c2)
        print("Shape of expected matrix:", c1.shape)
        print("Shape of obtained matrix:", c2.shape)
        print("Max. absolute error:", diff)
        raise ValueError(err)


def check_complex(m):
    """Check conversion of ``QStateMatrix`` to complex matrix
    
    It test the methods for conversion of a matrix to a 
    complex matrix 8with and without previous reduction) against
    the python function ````QStateMatrix()``.    
    """
    c0 = slow_complex(m)
    c1 = m.copy().complex()
    err = "Complex matrix from unreduced state"
    compare_complex(c0, c1, err, m)


#####################################################################
# Create radnom indices
#####################################################################

def random_slice(length):
    """Return data for indexing a random slice of a vector

    Let ``v`` be a one-dimensional numpy array of a given ``length``.
    
    We return a random triple ``(start, stop, step)`` for indexing a 
    the array ``v`` in the form ``v[start:stop:step]``.
    """
    start = randint(-length, length-1)
    stop = randint(-length, length-1)
    down = stop % length < start % length
    step = max(1, randint(0, abs(start - stop)//2))
    step *= (-1)**down 
    return start, stop, step    

#####################################################################
# Display and check some matrices
#####################################################################




qs_matrix_data = [
     (0, 0, (1,0), []),
     (0, 0, (-3,6), 0),
     (3, 0, (-3,5), 5),
     (0, 6, (9,2), 25),
     (3,1, (1,0), []),
     (2, 2, (0,0), [0b110_10_01, 0b101_01_11, 0b011_01_00]),
     (2, 0, (0,0), [0b000_10, 0b010_01, 0b100_01]),
    ]


def create_display_testvectors():
    """yield instances of class ``QStateMatrix`` for tests """
    for rows, cols, factor, data in qs_matrix_data:
        m = QStateMatrix(rows, cols, data)
        m.mul_scalar(*factor) 
        yield m
    yield qs_unit_matrix(4) 
    for i in range(20):
        yield qs_rand_matrix(2, 0, 3)
    for i in range(20):
        yield qs_rand_matrix(1, 0, 2)
    for rows in range(6):
        for cols in range(5):
            for nr in [1,2] + list(range(rows+cols, rows+cols+3)):
                m = qs_rand_matrix(rows, cols, nr)
                m.mul_scalar(randint(-8, 8), randint(0,7))  
                yield m                
                


@pytest.mark.qstate
def test_qs_matrix(verbose = 0):
    """Basic test for class ``QStateMatrix`` 
    
    It tests the ``__str__ `` method, the conversion to a
    complex matrix and the echelonization and the reduction of 
    an instance of class ``QStateMatrix``.
    """
    FORMAT_REDUCED = False
    display_len =  len(qs_matrix_data)
    for ntest, m in enumerate(create_display_testvectors()):
        if verbose or ntest < display_len:
            print("TEST %s" % (ntest+1))
            print(m.copy())
        check_complex(m)
        m2 = m.copy().echelon()
        check_echelon(m2)
        check_eq_qstate(m,m2, "Echelonized")
        if verbose:
            print("Echelonized")
            print(m2)
        m2 = m.copy().reduce()
        check_echelon(m2, reduced = True)
        check_eq_qstate(m,m2, "Reduced")
        if verbose: 
            print("Reduced")
            print(m2)
    FORMAT_REDUCED = True

 

@pytest.mark.qstate
def test_getitem(verbose = 0):
    FORMAT_REDUCED = False
    for ntest, m in enumerate(create_display_testvectors()):
        if verbose:
            print("TEST %s" % (ntest+1))
            print(m)
        c = m.complex()
        f0, t0, d0 = random_slice(1 << m.shape[0])
        f1, t1, d1 = random_slice(1 << m.shape[1])
        c1_ref = c[f0:t0:d0, f1:t1:d1] 
        c1 =  m[f0:t0:d0, f1:t1:d1] 
        s = "index m[%d:%d:%d, %d:%d:%d]" % (f0, t0, d0, f1, t1, d1)
        if verbose:
            print(s)
            print(c1)
        compare_complex(c1_ref, c1, s, m) 
        
        i0 = randint(-1 <<  m.shape[0], (1 <<  m.shape[0]) - 1 )     
        i1 = randint(-1 <<  m.shape[1], (1 <<  m.shape[1]) - 1 ) 
        c1_ref = c[f0:t0:d0, i1] 
        c1 =  m[f0:t0:d0, i1] 
        s = "index m[%d:%d:%d, %s]" % (f0, t0, d0, i1)        
        if verbose:
            print(s)
            print(c1)
        compare_complex(c1_ref, c1, s, m)  

        c1_ref = c[i0, f1:t1:d1] 
        c1 =  m[i0, f1:t1:d1] 
        s = "index m[%d, %d:%d:%d]" % (i0, f1, t1, d1)
        if verbose:
            print(s)
            print(c1)
        compare_complex(c1_ref, c1, s, m)  

        c1_ref = c[i0, i1] 
        c1 =  m[i0, i1] 
        s = "index m[%d, %d]" % (i0, i1)
        if verbose:
            print(s)
            print(c1)
        compare_complex(np.array(c1_ref), np.array(c1), s, m)  
    FORMAT_REDUCED = True

 
#####################################################################
# Test the generation of a monomial matrix
#####################################################################


def eval_nonomial(data):
    """Auxiliary function for function ``slow_complex``.
    
    Yet to be documented
    
    TODO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    This function sets map(0b000..01) to data[0] ^ data[1]
    The C function might set  map(0b000..01) to data[1]
    
    Find a reasonable definition of this function!!!!!!!!!!!!!!!
    """
    nqb = len(data) - 1
    res = np.zeros( (1 << nqb, 2), dtype = np.int32)
    data = data[:]
    mask = (2 << nqb) - 1
    data[0] &= mask
    for i in range(1, nqb + 1):
        data[i] &= mask
        mask += mask + 1
        
    mask =  (1 << nqb) - 1       
    for j in range(1 << nqb):
        v = j + j + 1    
        a = 0
        for i, d in enumerate(data): 
            a ^= d & -((v >> i) & 1)
        q = (a >> nqb) & v        
        for sh in [32, 16, 8, 4, 2, 1]:
            q ^= q >> sh 
        res[j,0] = a & mask
        res[j,1] = 1 - 2*(q & 1)
    return res


def a_binary(data, length):
    s = ""
    for n in data:
        n &= ((1 << length) - 1)
        b = format(n, "b")
        s +=  "0" * (length - len(b)) + b + "\n"
    return s


      


qs_monomial_matrix_data = [
     [1],
     [0b10, 0b00],
     [0b11, 0b11],
    ]
 
def rand_monomial_data(nqb):
    k = 1 << (2 * nqb)
    return [randint(0, k-1) for i in range(nqb+1)]
 
def create_monomial_testvectors():
    """yield instances of class ``QStateMatrix`` for tests """
    for data in qs_monomial_matrix_data:
        yield data
    for nqb in range(8):
        for k in range(3):
            yield(rand_monomial_data(nqb))
            
 
@pytest.mark.qstate
def test_monomial(verbose = 0):
    FORMAT_REDUCED = False
    for ntest, data in enumerate(create_monomial_testvectors()):
        nqb = len(data) - 1
        m = qs_column_monomial_matrix(data)
        if verbose:
            print("TEST %s, nqb = %d" % (ntest+1, nqb))
            s_data =  a_binary(data, 2*nqb)
            print("input:\n" + s_data)
            print(m)

        c = m[:,:].T
        s = np.sum(abs(c), axis = 1) - 1
        error = np.amax(abs(s)) > EPS
        if (verbose or error):
            if error:
                err = "Error in checking monomial matrix"
                print("\n" + err + "\n")
            s_data =  a_binary(data, 2*nqb)
            print("data  = \n" +  s_data)
            print("obtained:", m)
            if nqb <= 3:
                print("obtained matrix\n\n", c)
            print("Shape of obtained matrix:", m.shape, c.shape)
            if error:
                print("Max. absolute error:", np.amax(abs(s)))
                raise ValueError(err)
        res = eval_nonomial(data)
        for j, w in  enumerate(res):
            k = w[0]
            value = w[1]
            assert c[j,k] == value, (nqb, j, k, c[j,k], value, res)
        # Test qs_rown_monomial_matrix(data)
        m_r = qs_row_monomial_matrix(data)
        #print("m_r =", m_r, "m_c=", m ," m_T =", m.T)      
        assert m_r == m.T, (str(m_r), str(m.T))
    FORMAT_REDUCED = True


#####################################################################
# Test some error cases
#####################################################################
  
@pytest.mark.qstate
def test_qs_errors(verbose = 0):
    FORMAT_REDUCED = False
    with pytest.raises(ValueError):
        QStateMatrix(10000, 10000) 
    with pytest.raises(ValueError):
        QStateMatrix(1, 1, [0, 0, 0xf]) 
    m1, m2 = qs_unit_matrix(1),  qs_unit_matrix(2)
    with pytest.raises(ValueError):
        m1 @  m2
    with pytest.raises(ValueError):
        m1 *  m2
    m =  QStateMatrix(3, 3, [0, 0, 0xf], 2) 
    with pytest.raises(ValueError):
       m.power(-1) 
    FORMAT_REDUCED = True






import sys
import os

sys.path.insert(0, os.path.abspath("src"))
import mmgroup

import pytest


from mmgroup import MMSpace, characteristics, INT_BITS, MMV, MM0

#space = dict()
#for p in characteristics():
#    space[p] = MMSpace(p)
group = MM0

ITERATIONS = 10000
ITERATIONS_NOBREAK = 1000000



def bench(p, operation = [], iterations = ITERATIONS, break_g = True):
    v = MMV(p)('R')
    #print(operation)
    g = MM0(operation)
    v.mul_exp(g, iterations, break_g = break_g)
    return v.last_timing, iterations


def bench_nobreak(p, operation = [], iterations = ITERATIONS_NOBREAK):
    return bench(p, operation, iterations, break_g = False) 

def quot(f, *args):
    q, r = f(*args)
    return q/r

def quot_ms(f, *args):
    return "%9.6f" % (1000 * quot(f, *args))





def benchmark():
    print("""
Benchmarking monster operations a %d-bit system.
All times are given in milliseconds.
""" % (INT_BITS)
    )    
    for p in characteristics():    
        print("Operation modulo", p)
        op = [('p', 22), ('d', 127)]
        print ("p    ", quot_ms(bench_nobreak, p, op), " local optimization")

        print ("p    ", quot_ms(bench, p, op))
        op = [('p', 23), ('d', 12745645)]
        print ("p odd", quot_ms(bench, p, op))

        op = [('x', 1237), ('y', 567),]
        print ("xy   ", quot_ms(bench, p, op))

        op = [('l', 2)]
        print ("l    ", quot_ms(bench, p, op))

        op = [('t', 2)]
        print ("t    ", quot_ms(bench, p, op))

        op = [('t', 2), ('p', 23), ('d', 12745645),
            ('x', 1237), ('y', 567),
            ('l', 1),  ('p', 442),
            ('l', 1),  ('p', 142345345),
            ('l', 1),  ('p', 44865756), ('y', 567),
        ]

        print ("elem ", quot_ms(bench, p, op))


        op = [('t', 2), ('p', 23), ('d',12745645),
            ('x', 1237), ('y', 567),
            ('l', 1)
        ]
        print ("rand ", quot_ms(bench, p, op))
        print("")


@pytest.mark.very_slow
@pytest.mark.slow
@pytest.mark.bench
@pytest.mark.user
def test_benchmark():
    benchmark()


if __name__ == "__main__":
    benchmark()





"""Abstract model of a representation the monster group

Let R_p be the 196884-dimensional representation 196884x of the monster 
group modulo a small odd number p as described in [Seys19]. Class
AbstractMmRepSpace is an abstract class for modelling representation
R_p. A small odd modulus p < 256 is passed to the constructor of that
class as well as a class modelling the monster group Mm.

The user should refer to the concrete class mmgroup.MMSpace instead.

Class AbstractMmRepSpace handles the input and output of vectors 
in R_p. Here a basis vector of R_p is constructed from a tuple 
(tag, i0, i1), where the tag is a single capital letter as described 
in file mm_aux.c, and i0 and i1 are integers describing the basis 
vector.

Method unit() of class AbstractMmRepSpace constructs a basis vector
from a tuple (tag, i0, i1j). Standard operators "+" and "*" can be 
used for vector addition and scalar multiplication, and also for 
right multiplication of a vector by a group element. Optionally, a 
tuple (scalar, tag, i0, i1) can be passed with 'scalar' an integer 
or a specific small letter indicating a random scalar.

Since class AbstractMmRepSpace is a subclass of class AbstractRepSpace,
vectors of R_p can also be constructed by calling an instance of a
suitable subclass of class AbstractMmRepSpace with arguments describing 
the vector to be constructed. See method __call__ in class 
AbstractRepSpace for details.

The monster group Mm acting on R_p by right multiplication is modelled
by an instance of subclass of class mmgroup.structures.AbstractGroup.
Such a class has methods for generating elements of Mm. The operation 
of x in Mm on v in R_p is simply coded as v * x.

Class AbstractMmRepSpace is an abstract class that supports two
implementations of the concrete class mmgroup.MMVector. 


Internal operation
------------------

The representation (tag, i0, i1) of a basis vector of R_p is human
readable and in one-to-one correspondence with the description of 
basis vectors of R_p in [Seys19]. On the C level, dealing with
tuples and characters causes unnecessary overhead. Here we 
represent a scalar multiple of a basis vector in the so-called
sparse format. This means that a multiple of a basis vector is
represented as an unsigned 32-bit integer, where certain bit fields
of the integer correspond to the parts (scalar, tag, i0, i1) of the 
tuple, and another bit field corresponds to the coordinate.
Details are given in the documentation of the module mm_aux.c.
Vectors in R_p are can also be represented as arrays of 32-bit
integers. That representation of a vector is called the 'sparse'
representation. 

In python, a sparse representation of a vector in R_p is implemented
as a one-dimension numpy array of dtype np.uint32. 

Class AbstractMmRepSpace handles the operations on vectors an group
elements. It makes no specific assumptions on the internal 
representation of a vector or a group element. Thus it can easily be 
adjusted to a fast C implementation of the group operation on vectors.

However, class  AbstractMmRepSpace assumes that conversion between 
the internal format and the sparse format of a vector is supported, 
e.g. by fast C routines. It handles the conversion between the sparse
representation of a vector and the representation of a vector as a
list of tuples. This includes the implementation of the usual methods
__getitem__ and __setitem__ for vectors.

Caution!
v['A', i, j] and v['A', j, i] refer to the same coordinate of 
vector v. Item assignment is done in the same way as in numpy. 
Thus the result of the following assignment is not well defined:

    v['A',1:3,1:3] = [[11,12], [21,22]]

Depending on the internal order of the item assignment it may set
v['A',1,2] and v['A',2,1] either to 12 or to 21. A similar problem 
may occur with tags 'B' and 'C'.


Subspaces spanned by basis vectors
----------------------------------

We also use the sparse representation to describe subspaces of R_p 
generated by basis vectors. Function sparse_subspace() creates a 
numpy array describing such a subspace. Note that the
characteristic p need not be known for describing such a subspace.

Subclasses of class AbstractMmRepSpace
--------------------------------------

A subclass of class AbstractMmRepSpace must implement the following
methods:

zero:
should return the zero vector

iadd:
vector addition, see class AbstractRepSpace for details.

imul_scalar:
scalar multiplication, see class AbstractRepSpace for details.

imul_group_word:
multiply a vector with an atom of the group Mm,  see class 
AbstractRepSpace for details. See class MGroupN for decomposing
an element of Mm into atoms.

reduce:  
Reduce the representation of a vector to a standard form. This is  
not necessary if each vector has a unique internal representation.

as_sparse:
Converts a vector (which is an instance of a subclass of class)
AbstractRepVector to its sparse representation. This function 
should return a numpy array containing the sparse representation.


getitems_sparse:
Support for reading components of a vector with __getitem__.
This method takes an array of integers containing the descriptions
of the basis vectors for which the co-ordinates are to be read.
Here the basis vectors must be given in sparse format.
In that array, the function must update the fields containing 
the coordinates.

setitems_sparse:
Support for writing components of a vector with __setitem__.
This method takes an array of integers containing the descriptions
of the basis vectors for which the co-ordinates are to be written,
and the co-ordinates.
Here the basis vectors must be given in sparse format.


additems_sparse:
Support for item assignment on a vector with __setitem__.
This method takes an array of integers containing the descriptions
of a vector v in sparse format. It add vector v to a given vector.


equal_vectors:
Check if two vectors are equal.

optional methods:

set_rand_uniform:
randomize a vector a uniform distributed


"""

#234567890123456789012345678901234567890123456789012345678901234567890

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import warnings
import copy
import numbers
import re
from numbers import Integral
from random import randint, randrange, sample
from functools import partial
from collections import defaultdict
import numpy as np


from mmgroup.structures.abstract_rep_space import mod_rand_invertible
from mmgroup.structures.abstract_rep_space import mod_rand_unit
from mmgroup.structures.abstract_rep_space import AbstractRepSpace
from mmgroup.structures.abstract_rep_space import AbstractRepVector


from mmgroup.structures.parse_atoms import AtomDict 
from mmgroup.structures.parse_atoms import eval_atom_expression, ihex 


from mmgroup.structures.mm_space_indices import tuple_to_sparse
from mmgroup.structures.mm_space_indices import sparse_from_indices
from mmgroup.structures.mm_space_indices import sparse_add_data
from mmgroup.structures.mm_space_indices import sparse_to_ndarray
from mmgroup.structures.mm_space_indices import sparse_to_str
from mmgroup.structures.mm_space_indices import sparse_to_tuples
from mmgroup.structures.mm_space_indices import purge_sparse_entry






ERR_MM_CONV = "Sparse representation of MM vectors not fully supported"


try:
    from mmgroup.mm import mm_aux_mul_sparse
except (ImportError, ModuleNotFoundError):
    warnings.warn(ERR_MM_CONV, UserWarning)   


#######################################################################
# Import derived classed
#######################################################################

import_pending = True

def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending, XLeech2, mm_aux_index_leech2_to_sparse
    from mmgroup.mm import mm_aux_index_leech2_to_sparse 
    from mmgroup.structures.xleech2 import XLeech2
    import_pending = False


#######################################################################
# class AbstractMmRepVector
#######################################################################



def xleech2_to_sparse(p, v):
    if import_pending:
        complete_import()
    assert isinstance(v, XLeech2)
    v2 = v.value
    sp = mm_aux_index_leech2_to_sparse(v2)
    if sp == 0:
        err = "Instance of class XLeech2 is not of type 2"
        raise ValueError(err)
    sign = p-1 if v2 & 0x1000000 else 1
    return np.array([sp + sign], dtype = np.uint32)


class AbstractMmRepVector(AbstractRepVector):
    space = None # Vector space associated with vector type

    def __init__(self, p, tag = 0, i0 = None, i1 = None):
        self.set_zero(self, p)
        add_vector(self, tag, i0, i1)

    def set_zero(self, p = 0):
        """set a vector to zero

        The vector is to be interpreted as a vevtor of integers
        modulo ``p``. In case ``p == 0`` it is to be inerpreted
        as a vector of rational numers
        """
        raise NotImplementedError("Abstract method")


    def as_bytes(self):
        """Return vector as an array of bytes.

        The function returns a one-dimensional numpy array of ``196884``
        unsigned integers of bit length ``8``.

        For a vector ``v`` the expression ``v.as_bytes()`` is equivalent 
        to ``v['E':]``.
        """
        return self.space.as_bytes(self)

    def as_sparse(self):
        """Return vector in sparse representation

        The sparse representation of a vector is returned as a 
        one-dimensional numpy array of 32-bit integers. Here each 
        entry of that array represents a nonzero component of
        the vector. Such an entry it interpreted as described
        in method ``from_sparse`` of class |MMSpace|.

        Entries with value zero are dropped.
        """
        return self.space.as_sparse(self)

    def as_tuples(self):
        """Return vector in tuple representation

        The function returns a list of tuples ``(factor, tag, i0, i1)``.
        Here ``(tag, i0, i1)`` describes a basis vector and ``value``
        is the coordinate of that vector.

        Entries with coordinate zero are dropped.
        """
        return self.space.as_tuples(self)

    def get_sparse(self, a_sparse):
        """Get coordinates of vectors in sparse representation.

        Here 'a_sparse' is an array-like object containing the
        indices of basis vectors to be read in sparse representation. 

        The function returns a numpy array 'a' in sparse representation
        with the indices of the basis vectors read from the input 
        'a_sparse'. It updates the corresponding values in 'a'
        a with the coordinates of the basis vector taken from v. 
        """
        a = np.array(a_sparse, dtype = np.uint32)
        a1 = a.reshape(-1)
        self.space.getitems_sparse(self, a1)
        return a

    def set_sparse(self, a_sparse):
        """Set coordinates of vectors given by sparse representation.

        Here 'a_sparse' is an array-like object containing the
        indices of the basis vectors and the values to be written
        in sparse representation. 

        The function updates the coordinates of the vector given
        by the entries of 'a_sparse' with the corresponding values.

        The order of these write operations is unspecified. Therefore
        array 'a_sparse' should not contain any duplicate basic 
        vector indices with inconsistent values. 
        """
        a = np.array(a_sparse, dtype = np.uint32)
        a1 = a.reshape(-1)
        self.space.setitems_sparse(self, a1)

    def projection(self, *args):
        """Return projection of the vector ``v`` onto a subspace.

        Each argument describes a subspace of this space ``V``. 
        The projection of the vector to the space 
        ``W`` spanned by  all these subspaces is returned.

        In the current version each argument must be a tuple
        ``(tag, i0, i1)`` that describes a subspace generated by a
        basis vector. A legal ``tag`` is letter in the string
        ``ABCTXYZD``, as explained in table :ref:`table-vector-tags` 
        and class |MMSpace|. 
        """
        a = sparse_subspace(*args)
        self.space.getitems_sparse(self, a)
        v = self.space.zero(self.p)
        self.space.setitems_sparse(v, a)
        return v
        
    def raw_str(self):
        return self.space.raw_str_vector(self)
        


######################################################################
# class AbstractMmRepSpace
######################################################################


class AbstractMmRepSpace(AbstractRepSpace):
    """Models the 198884 dimensional representation the monster group. 

    A basis vector is modeled as a tuple  vtype = (tag, i0, i1)  as
    described in file mm_aux.c. 'tag' is a capital letter. 'i0' and
    'i1' are two integers or small letters indicating a random
    number. A scalar multiple of a basis vector may be written as 
    (scalar, tag, i0, i1) where 'scalar' is an integer or a small 
    letter indicateing a random scalar.

    Method self.unit(scalar, tag, i0, i1) or self.unit(tag, i0, i1) 
    can be used to construct a (scalar multiple of a) unit vector.

    Here 'tag' is a single capital letter representing the tag 
    of the unit vector, and i0 and i1 are integers.

    Valid tags are 'A', 'B', 'C', 'T', 'X', 'Y', 'Z' as described       
    in mm_aux.c, and also 'D' with 'D_i' meaning 'A_i_i'.
      
    The optional the expression 'scalar' may be an integer 
    denoting a scalar factor to be multiplied with the unit
    vector or one of following letters for generating a random 
    multiple of a unit vector:
      
      'r'   generate a random multiple of a unit vector which may  
            be a zero or a zerodivisor
      'n':  generate a random multiple of a unit vector which is  
            not a zerodivisor
      's'   generate a (possibly negated) unit vector  
      'u'   generate a unit vector 

    i0 or i1 may also be the string 'r' indicating a random value.

    Scalars are integers modulo a small odd number p given in 
    the constructor.   Operations  + and - are supported for vectors.
    Operation and *, /, << and >> are supported for vectors and 
    integers modulo p.  Here 1/b is the modular inverse of b,  a << b 
    means a * 2**b,  a >> b means a * 2**(-b).

    Each instance of this class has a group operating on it, which 
    is passed in the constructor. Such a group should be an instance 
    of class mmgroup.structures.MGroupN. Elements of the group operate 
    on vectors by right multiplication. Division of  a vector by a 
    group element means multiplication with the inverse element.
    
    Vectors in a space may be created by calling an instance of this
    class representing that space with any number of arguments. All
    these arguments are added up, and each argument v may be one of 
    the following.
    
    - v may be a string which is a valid python expression.
      Here identifiers of the following format are recognized as unit
      vectors:
      
         <scalar>_<tag>_<i0>_<i1>   or   <tag>_<i0>_<i1> .

      Here letters and integers are interpreted as in method
      self.unit(). Integers may be decimal, hexadecimal 0xH...H as 
      in C or of the form DX...Xh with D a decimal and X a 
      hexadecimal digit. 
   
      Addition, subtraction and scalar multiplication may be encoded
      in the python expression to generate more general vectors.
      
    - v may be a vector in that space.
      
    - v may be an element of another space V.
      If there is a homomorphism phi from V to this space 
      (set by  method set_preimage) this evaluates to phi(v).

    - v may be a tuple (tag, i0, i1) or (scalar, tag, i0, i1).
      Here scalar, tag, i0, and i1 are interpreted as in
      method self.unit().

    - v may be the integer 0. This means the zero vector.
    """

    space_name = "AbstractMmRepSpace"

    vector_type = AbstractRepVector


    #######################################################################
    # Constructor
    #######################################################################


    def __init__(self):
        """Create a 196884-dimensional representation of the monster

        All calculations are done modulo the odd number 3 <= p < 256.
        'group' is the group operating on the vector space by right
        multiplication. This should be an instance of class 
        mgroup_n.MGroupN.
        """
        pass

    #######################################################################
    # Obtaining and setting components via sparse vectors
    #######################################################################

    def getitems_sparse(self, v, a_sparse):
        """Obtain selected components of a vector 
        
        Support for reading components of a vector v with __getitem__().

        This method takes a numpy array 'a_sparse' of integers of dtype 
        numpy.uint32 containing the descriptions of the basis vectors for 
        which the co-ordinates are to be read.  Here the basis vectors 
        must be given in sparse format. In that array, the function must 
        update the fields 'scalar' in each entry of the array with the 
        coordinate of the vector  v  corresponding to the basis vector
        given be the fields  (tag, i0, i1). 

        Here vector v is a standard vector in this space.
        """
        raise NotImplementedError("Abstract internal method") 

    def setitems_sparse(self, v, a_sparse):
        """Update selected components of a vector 

        Support for writing components of a vector v with __setitem__().

        This method takes a numpy array 'a_sparse' of integers of dtype 
        numpy.uint32 containing the descriptions of the basis vectors for 
        which the co-ordinates are to be written. The array must also
        contain the values to be written. Here the basis vectors and
        their values must be given in sparse format. The function must 
        update the corresponding coordinates of the vector v.

        Caution:
        If duplicate coordinates are given in a_sparse, these coordinates  
        may be updated in any order. The user is responsible for not 
        assigning different values to the same co-ordinate.

        Here vector v is a standard vector in this space.
        """
        raise NotImplementedError("Abstract internal method") 


    def additems_sparse(self, v, a_sparse):
        """Add a vector in sparse representation to vector v.

        This method takes a numpy array 'a_sparse' of integers of dtype 
        numpy.uint32 containing the description of a vector v2 in sparse 
        representation. It computes 

             v  =  v + v2 .

        Here vector v is a standard vector in this space.
        """
        raise NotImplementedError("Abstract internal method") 


    #######################################################################
    # Conversion from and to sparse representation 
    #######################################################################


    def from_sparse(self, a_sparse):
        """Return vector a given by its sparse representation.

        Here ``a_sparse`` is an array-like list of unsigned 
        integers that contains the sparse representation of a vector. 
        Each  entry of that array represents a component of
        the vector. Such an entry it interpreted as follows:

        .. table:: Bit fields in an entry of a sparse representation
          :widths: 28 24 24 24 24 
 
          ========= =========== =========== ========== ========== 
          Component ``tag``     ``i0``      ``i1``     ``factor``
          Bits      ``27 - 25`` ``24 - 14`` ``13 - 8`` ``7 - 0``
          ========= =========== =========== ========== ==========

        This corresponds to the tuple ``(factor, tag, i0, i1)``
        which denotes a multiple of a basis vector as described in
        the section :ref:`mmrep-label`. 

        Tags are mapped to integers as follows:

        .. table:: Mapping of integers to tags
          :widths: 14 14 14 14 14 15 15

          ======= ======= ======= ======= ======= ======= ======= 
          ``'A'`` ``'B'`` ``'C'`` ``'T'`` ``'X'`` ``'Z'`` ``'Y'``
           ``1``   ``2``   ``3``   ``4``   ``5``   ``6``   ``7``
          ======= ======= ======= ======= ======= ======= ======= 


        The array-like object ``a_sparse`` is interpreted in the same
        way as in the construction of a numpy array of unsigned
        ``32``-bit integers.

        .. warning::
          If duplicate coordinates are given in ``a_sparse``, these 
          coordinates may be updated in any order. The user is responsible 
          for not assigning different values to the same coordinate.
        """
        v = self.zero()
        self.setitems_sparse(v, a_sparse)
        return v
    

    def as_sparse(self, v1):
        """Return sparse representation of vector v1
        
        This should return the sparse representation of a vector as
        a list of integers. The list is not sorted.
        
        The function may be implemented as a generator.
        """
        raise NotImplementedError("Abstract method")


    def via_sparse(self, v1):
        """Conversion from another vector space via sparse rep"""
        raise NotImplementedError
        assert v1.space.p % self.p == 0
        sp = v1.space.as_sparse(v1)
        sp = (sp & 0xffffff00) + (sp & 0xff) % self.p
        return self.from_sparse(sp)
        
    #######################################################################
    # Conversion to a list of tuples 
    #######################################################################

    def as_tuples(self, v1):
        """Convert vector v1 from sparse representation to list of tuples

        The function returns a list of tuples

           (scalar, tag, i0, i1)

        Here (tag, i0, i1) represents a unit vector and the integer
        'scalar' is the coordinate of that unit vector.
        """
        a_sparse = self.as_sparse(v1)
        return sparse_to_tuples(a_sparse) 

    #######################################################################
    # Creating vectors 
    #######################################################################

        
    def unit(self, *data):
        """Return unit vector given by ([scalar], tag, i0, i1)

        Here (possibly a random multiple of) a basis vector is
        created according to the input parameters 
        ([scalar], tag, i0, i1). 
    
        The 'scalar' is optional and defaults to 1.
        """        
        raise NotImplementedError
        return  self.from_sparse(tuple_to_sparse(self.p, *data))   


    def rand_vector(self, tags = None, min_weight=1, max_weight=0):
        """return a random vector

        'tags' is a string as in method unit(). The vector contains 
        between 'min_weight' and 'max_weight' nonzero components, 
        as returned by method rand_unit().
        In is possible that such components may cancel.

        If no argument is given, we return a uniform random vector, 
        if the class supports this.
        """
        raise NotImplementedError
        if tags is None:
            return self.rand_uniform()
        n = randint(min_weight, max(min_weight,max_weight))
        v = self.zero()
        for i in range(n):
            tag = sample(tags, 1)[0]
            v += self.unit(tag, "r")
        return v

    def rand_uniform(self):
        """Create a random vector with uniform distribution"""
        raise NotImplementedError("Cannot generate a uniform random vector") 

    def set_rand_uniform(self, vector):
        """Create a random vector with uniform distribution"""
        raise NotImplementedError("Cannot generate a uniform random vector") 


    def iadd_tuple(self, v1, t):
        """Compute v += self.unit(*t), return v"""
        raise NotImplementedError
        v2 = tuple_to_sparse(v1.p, *t)
        return self.additems_sparse(v1, v2)

    #######################################################################
    # getitem and setitem
    #######################################################################


    def vector_get_item(self, v1, item):
        assert v1.space == self
        if not isinstance(item, tuple):
            item = (item,) 
        shape, a_sparse = sparse_from_indices(v1.p, *item)
        self.getitems_sparse(v1, a_sparse)
        return sparse_to_ndarray(v1.p, shape, a_sparse)

    def vector_set_item(self, v1, item, value):
        assert v1.space == self
        if not isinstance(item, tuple):
            item = (item,) 
        shape, a_sparse = sparse_from_indices(v1.p, *item)
        if isinstance(value, AbstractRepVector):
            self.getitems_sparse(value, a_sparse)
        else:
            sparse_add_data(v1.p, shape, a_sparse, value)
        self.setitems_sparse(v1, a_sparse) 

         

    #######################################################################
    # Formatting a vector 
    #######################################################################

    


    def str_vector(self, v1):
        s = sparse_to_str(v1.p, self.as_sparse(v1))
        return "%s<%d;%s>" % (self.space_name, v1.p, s)
           

    def raw_str_vector(self, v1):
        return sparse_to_str(v1.p, self.as_sparse(v1))


######################################################################
# Creating a sparse representation of a subspace
######################################################################


def sparse_subspace(*args):
    """Create a sparse representation of a subspace of MM vectors
    
    The function returns the description of a subspace of the space 
    R_p spanned by basis vectors of R_p. That description is returned
    as a sorted numpy array of type uint32. Each entry of that array 
    describes a basis vector in sparse representation (with component 
    'value' set to zero). 

    Each argument describes a subspace of R_p; and the span of all 
    these subspaces is returned. An argument may be:

    - A set or an array-like object containing integers
        Then all entries are interpreted as sparse representations
        of basis vectors, ignoring the scalar factor. Integers not
        representing basis vectors correctly are ignored.
        This feature is not implemented in the current version!!!

    - A tuple (tag, i0, i1) representing a basis vector.
        Here a tuple is interpreted as a vector in the same way as 
        in a call to an instance of class AbstractMmRepVector. 
        Note that such a call is used for generating a vector.

    - A vector v in a vector space R_p
        Here the vector space R_p must be an instance of a 
        (subclass of) class AbstractMmRepVector, and v must be a
        vector in that space. Then v describes the subspace spanned 
        by all basis vectors where the coordinate of v is not zero. 
        This feature is not implemented in the current version!!!
    """
    a = [np.empty(0, dtype = np.uint32)]
    for v in args:
        if isinstance (v, set):
            v = list(v)
        if isinstance(v, tuple):
            a.append(tuple_to_sparse(1, *v))
        elif isinstance(v, AbstractMmRepVector):
            err = "MM Vector input in function sparse_subspace not supported"
            raise(NotImplementedError, err)
            a.append(v.as_sparse())
        else:
            err = "Array input in function sparse_subspace not supported"
            raise(NotImplementedError, err)
            va = np.array(v, dtype = np.uint32).reshape(-1)
            for i, x in enumerate(va):
                va[i] = purge_sparse_entry(x)
            a.append(va)
    a_out = np.unique(np.concatenate(a) & 0xffffff00)
    if len(a_out) and a_out[0] == 0:
        a_out = a_out[1:]
    return a_out


######################################################################
# Constructing a vector
######################################################################


FRAME = re.compile(r"^([A-Za-z_])+\<([0-9]+;)?(.+)\>$") 

MMVector = None

def parse_mm_space_string(space, p, s):
    global MMVector
    m = FRAME.match(s)
    p_intern, space_intern = None, None
    if m:
        space_name, p_str, string = (m[1], m[2], m[3])
        if p_str: p_intern = int(p_str[:-1])
        space_intern = SPACES_BY_NAME.get(space_name)
    else:
        space_intern = space.vector_type
        string = s
        p_intern = p
    if p_intern is None:
        p_intern = p
    if space_intern is None:
        if MMVector is None:
            from mmgroup import MMVector
        space_intern = MMVector
    f = AtomDict(partial(space_intern, p_intern))
    return eval_atom_expression(string, f)
    



ERR_MMV_TYPE = "Cannot convert type '%s' object to MM vector"


def add_conv_vector(vector, tag, factor = 1):
    p = vector.p
    space = vector.space
    if isinstance(tag, str):
        a = parse_mm_space_string(space, p, tag)
    else:
        a = tag
    if (a.space == space and a.p == p):
        factor %= p
        if factor <= 1:
            if factor: space.iadd(vector, a)
        else:
            space.imul_scalar(a, factor)
            space.iadd(vector, a)
    elif (a.space, space) in MM_VECTOR_CONVERSIONS:
        f = MM_VECTOR_CONVERSIONS[(a.space, space)]
        return f(p, factor, a)
    else:
        v = a.as_sparse()
        v1 = np.zeros(len(v), dtype = np.uint32)
        res = mm_aux_mul_sparse(a.p, v, len(v), factor, p, v1)
        if res < 0:
            err = "Cannot reduce sparse MM vector modulo %d"
            raise ValueError(err, p)
        space.additems_sparse(vector, v1[:res])

def add_vector(vector, tag = 0, i0 = None, i1 = None):
    if not tag:
        return
    p = vector.p
    space = vector.space
    if isinstance(tag, str) and len(tag) == 1:
        if tag in "ABCTXZYDEIJUS0":
            space.additems_sparse(vector,
                tuple_to_sparse(p, tag, i0, i1))
        elif tag == "R":
            space.set_rand_uniform(vector)
        elif tag == "V":
            v1 = space.from_bytes(p, i0)
            space.iadd(vector, v1)
        else:
            err = "Illegal tag '%s' for MM vector"
            raise ValueError(err % tag)
    elif isinstance(tag, (AbstractMmRepVector, str)):
         add_conv_vector(vector, tag, factor = 1)
    elif isinstance(tag, list):
        for x in tag:
            if isinstance(x, tuple):
                space.additems_sparse(vector,
                    tuple_to_sparse(p, *x))
            elif isinstance(x, (str, AbstractMmRepVector)):
                add_vector(vector, x)  
            elif x:
                raise TypeError(ERR_MMV_TYPE % type(x))                                                     
    elif isinstance(tag, Integral):
        add_conv_vector(vector, i0, factor = tag)
    else:
        if import_pending:
            complete_import()
        if isinstance(tag, XLeech2):
            space.additems_sparse(vector, (xleech2_to_sparse(p, tag)))
        else:
            raise TypeError(ERR_MMV_TYPE % type(tag))       

  
MM_VECTOR_CONVERSIONS = {
}


SPACES_BY_NAME = {
}









"""Auxiliary functions for writing tables into C program"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import types
import sys
import re
import os
import warnings
import itertools
import ast
from io import IOBase
from operator import __or__, __xor__, __and__
from numbers import Integral
from collections.abc import Iterable
from functools import partial
import numpy






if sys.version_info[0] >= 3:
    file = IOBase 
    long = int    


import mmgroup.generate_c.generate_functions 


from mmgroup.generate_c.generate_functions import isinteger
from mmgroup.generate_c.generate_functions import format_item
from mmgroup.generate_c.generate_functions import sizeof_format
from mmgroup.generate_c.generate_functions import make_table
from mmgroup.generate_c.generate_functions import py_doc_to_comment
from mmgroup.generate_c.generate_functions import format_line
from mmgroup.generate_c.generate_functions import line_as_comment
from mmgroup.generate_c.generate_functions import named_table
from mmgroup.generate_c.generate_functions import eval_codegen_args
from mmgroup.generate_c.generate_functions import eval_codegen_table
from mmgroup.generate_c.generate_functions import eval_codegen_for
from mmgroup.generate_c.generate_functions import eval_codegen_with
from mmgroup.generate_c.generate_functions import eval_codegen_else
from mmgroup.generate_c.generate_functions import SaveDictFor
from mmgroup.generate_c.generate_functions import safe_eval
from mmgroup.generate_c.generate_functions import direct_single_arg
from mmgroup.generate_c.generate_functions import prepend_blanks
from mmgroup.generate_c.generate_functions import indent_subsequent_lines
from mmgroup.generate_c.generate_functions import eval_codegen_join
from mmgroup.generate_c.generate_functions import UserDirective
from mmgroup.generate_c.generate_functions import UserFormat
from mmgroup.generate_c.generate_functions import built_in_formats

class TableGenerator(object):
    """Automatically generate a .c file and a .h file from a *source*
 
    Basic idea: We copy a *source* file to the .c file. The source 
    contains directives to enter precomputed tables, code  snippets 
    or constants into the .c file, and prototypes into the .h file.

    We copy the content of the *source* file directly to the .c
    file.   In the *source* file we also parse certain directives 
    and we replace them by automatically generated code snippets 
    in the .c file and in the .h file. 

    The arguments given to the constructor of this class specify
    the meaning of the directives.

    For a more detailed description of these directives see module
    ``make_c_tables_doc.py`` 

    :param tables:

        Here ``tables`` is a dictionary of with entries of shape:: 

             ``name`` : ``table`` .

        ``name`` is a string that is a valid python identifier.
        ``table`` is the python object to be referred by
        that name whenever an argument is evaluated as a
        python expression.

    :param directives:

        Here ``directives`` is a dictionary of shape::

            ``function_name`` : ``function``  .

        The function ``function`` is executed when a
        directive of shape::

             // %%function_name

        occurs as a directive in the *source* file. The arguments 
        following the directive are evaluated and passed to the 
        function ``function``. That function should return a
        string. This string is merged into the output .c file.

    :param verbose: 

        optional, may be set ``True`` for verbose console output
    """

    m_kwd =  re.compile(r"\s*//\s*\%\%(\w+)(\*|\b)(.*)?")
    block_kwds = set(["FOR", "IF", "WITH"])
    ILLEGAL_INSIDE_BLOCK = "%s directive is illegal inside a codegen block"
    is_terminal = True, # True: Error if cannot valuate %{..}  expression
                        # False: Let expression unchanged in that case

    def __init__(self, tables, directives={}, verbose = False):
        """Creates a Code generator with tables and directives
 
        """
        self.verbose = verbose
        self.tables = tables  # This is never changed
        self.names = {}       # Updated version of self.names
        self.reset_names()    # initialize self.names from self.tables
        self.exported_table_names = {}
        #  dictionary python_table_name : c_table_name
        # 'python_table_name' is a key of dictionary tables.
        # 'c_table_name' is the corresponding name of that table
        # or entry in C.

        # Enter user-defined directives into dictionary self.directives
        self.directives = {}
        for fname, f in directives.items():
            if not isinstance(f, UserDirective):
                try:
                    f = UserDirective(f)
                except:
                    print("\nError:")
                    print("Could not register directive %s for code generator\n" % fname)
                    raise
            f.set_code_generator(self)
            self.directives[fname] = f
         

        # Enter built-in directives into dictionary self.directives
        # Dictionary 'built_in_directives' of is shape
        #     name : <directive>.
        # Here <directives> is a member functions of this class 
        builtin_directives = ( {
             "TABLE":        self.table, 
             "USE_TABLE":    self.use_table,
             "EXPORT":       self.export_,
             "SET_EXPORT":   self.set_export,
             "EXPORT_KWD":   self.set_export_kwd,
             "EXPORT_TABLE": self.export_table,
             "GEN":          self.gen,
             "COMMENT":      self.comment,
             "PY_DOCSTR":    self.docstr,
             "PYX":          self.pyx,
             "FOR":          self.block_for,
             "WITH":         self.block_with,
             "JOIN":         self.do_join,
             "IF":           self.block_if,
        } )
        self.directives.update(builtin_directives)

        self.use_table_pending = False
        self.export_pending = 0
        self.pxd_export_pending = 0
        self.export_kwd = ""
        self.args = ()       # positional arguments passed to formatting
                             # function, not used for C files
        self.block_stack = [] # stack for loop keywords such as FOR, IF, ..

        self.source_name = None  # name of C skeleteon file
        self.current_line = ""


    def reset_names(self):
        """Reset dictionary self.names to its original value self.tables

        This means that all C names for tables generated by the
        %%TABLE directive will be lost.
        """
        self.names.clear()
        #names.update(safe_locals)
        self.names.update(built_in_formats)
        self.names.update(self.tables)
        self.names.update({
            "NAMES" : self.names,  
            "TABLES" : self.tables, 
        })
        self.adjust_names()

    def adjust_names(self):
        for _, entry in self.names.items():
            if isinstance(entry, UserDirective):
                 entry.set_code_generator(self)


    def _check_names(self):
        for _, entry in self.names.items():
            if isinstance(entry, UserDirective):
                 assert entry.tg == self


    def table(self, args, *_):
        """built-in function TABLE"""
        name, table, format_  = eval_codegen_table(self.tables, args)
        if name and self.C_table_name:
            self.names[name] = named_table(self.C_table_name, table)
            if self.C_table_export:
                self.exported_table_names[name] = self.C_table_name
        self.C_table_name = None
        self.C_table_export = False
        try:
            self.table_size_ += len(table) * sizeof_format(t_args[0])
        except:
            pass
        return make_table(table, format_), ""

    def use_table(self, args, *_):
        """built-in function USE_TABLE"""
        self.use_table_pending = 1 + self.export_pending
        self.export_pending = 0
        self.pxd_export_pending = 0
        return "", ""

    def export_table(self, args, *_): 
        """built-in function EXPORT_TABLE"""
        self.use_table_pending = 2
        self.export_pending = 0
        self.pxd_export_pending = 0
        return self.export_kwd, self.export_kwd


    def comment(self, args, *_):
        """built-in function COMMENT, deprecated!!"""
        warnings.warn(
            r"The %%COMMENT directive in the code generator is deprecated!",
            DeprecationWarning
        )
        return "", ""


    def export_(self, args, *_):
        """built-in function EXPORT"""
        self.export_pending = True
        self.pxd_export_pending =  'p' in args
        if self.pxd_export_pending and 'x' in args:
            self.pyx("<wrap pxi>")
        return self.export_kwd, self.export_kwd

    def set_export(self, args, *_):
        """built-in directive SET_EXPORT, deprecated!!!!"""
        err = "The SET_EXPORT directive is no longer supported"
        raise ValueError(err)


    def set_export_kwd(self, args, *_):
        """built-in function EXPORT_KWD"""
        self.export_kwd = args.strip() + "\n"
        return "", ""


    def gen(self, args, *_):
        """built-in function GEN"""
        if len(self.block_stack):
            raise TypeError(self.ILLEGAL_INSIDE_BLOCK % "GEN")
        arg = direct_single_arg(args, "GEN") 
        self.gen_h =  arg.find('h') >= 0
        self.gen_c =  arg.find('c') >= 0
        self.gen_c =  self.gen_c or not (self.gen_c or self.gen_h)
        return "", ""
        

    def pyx(self, args, *_):
        """built-in function PYX"""
        #args = format_line(args, self.names, self.args)
        self.pxd_entries.append("# PYX " + args.strip())
        return "", ""


    def docstr(self, args, *_):
        """built-in function PY_DOCSTR"""
        args= eval_codegen_args(self.tables, args)
        entry = args[0]
        if not isinstance(entry, str):
            entry = entry.__doc__
        if len(args) > 1 and args[1]:
            entry = format_line(entry, self.names, self.args, terminal=0)
        return py_doc_to_comment(entry), ""



    def parse_block(self, kwd, source):
        """General parser for reading a block from the source 

        Block must nested in the form:

        // %%KWD1
           ...
           // %%KWD2
           ...
           // %%END KWD2
           ...
        // %%END KWD1

        with indentation ignored. Keywords KWD1, KWD2 etc. are block
        keywords. The set of  block keywords is given by
        self.block_kwds.

        Once a directive with a block keyword has been parsed, a parser
        for reading the complete block corresponding to that block 
        keyword should be called. So in the example above, this parser
        reads the lines up to and including the final line
        '// %%END KWD1' of te block. Note that the first line
        '// %%KWD1' of the block has already ben read by the function
        calling this method. 

        'kwd' is the block keyword already read by the calling funtion.
        *source* is the iteratornthat provides the input lines.

        The function returns a triple

            (prefix_lines, code_lines, postfix_lines)

        of list of strings representing input lines. The lines in the
        list 'code_lines' form the body of the block, which will be
        processed for code generation by a subsequent function.

        'prefix_lines' and 'postfix_lines' are lists of comment lines 
        to be copied from the soure to the generated code file.

        Here in the standard case, 'prefix_lines' is empty and 
        'postfix_lines' contains the END statement of the block.
        """
        kwd_stack = [kwd]
        out_source = []
        for l in source:
            if self.verbose:
                level = len(self.block_stack) + len(kwd_stack) 
                print("%d. %s" % (level, l), end = "")
            out_source.append(l)
            m = self.m_kwd.match(l)
            if m:
                sub_kwd, _, args = m.groups()
                if sub_kwd in self.block_kwds:
                    kwd_stack.append(sub_kwd)
                elif sub_kwd == "END":
                    end_kwd = direct_single_arg(args, kwd = "END")
                    match_kwd = kwd_stack.pop()
                    if end_kwd != match_kwd:
                        kw = ("END " + end_kwd, match_kwd)
                        s =" % does not match % in codegen directive"
                        raise TypeError(s % kw) 
                    if len(kwd_stack) == 0:
                        return [], out_source[:-1], out_source[-1:]
        raise TypeError("Missing END %d directive in codegen" % kwd)
        
      

    def do_for_with_join(self, arg_list, source, save_names, c_out):
        """Special inner loop of FOR directive with JOIN prefix"""
        infix, suffix = self.join_pending
        for entry in arg_list:
            save_names.set_local(entry) # enter entry into self.names
            for c_out1, _ in self.iter_generate_lines(iter(source)):
                c_out.append(c_out1)
            if len(c_out) and  c_out[-1][-1:] == "\n":
                c_out[-1]  = c_out[-1][:-1] 
            c_out.append(infix)
        if len(c_out):
            c_out[-1] = suffix


    def block_for(self, args, quiet, source):
        """built-in function FOR"""
        local_vars, arg_list = eval_codegen_for(self.names, args)
        c_out, source, end_comment = self.parse_block("FOR", source)
        if quiet:
            c_out, end_comment = [], []
        h_out = []
        save_names = SaveDictFor(self.names, local_vars)
        self.block_stack.append("FOR")
        if self.join_pending:
            self.do_for_with_join(arg_list, source, save_names, c_out)
            self.join_pending = False
        else:
            for entry in arg_list:
                save_names.set_local(entry) # enter entry into self.names
                for c_1, h_1 in self.iter_generate_lines(iter(source)):
                    c_out.append(c_1)
                    h_out.append(h_1)
        self.block_stack.pop()
        save_names.restore()   # restore self.names
        return "".join(c_out + end_comment) , "".join(h_out)

    def block_with(self, args, quiet, source):
        """built-in function WITH"""
        local_vars, values = eval_codegen_with(self.names, args)
        c_out, source, end_comment = self.parse_block("WITH", source)
        if quiet:
            c_out, end_comment = [], []
        h_out = []
        save_names = SaveDictFor(self.names, local_vars)
        self.block_stack.append("WITH")
        save_names.set_local(values) # enter entry into self.names
        for c_1, h_1 in self.iter_generate_lines(iter(source)):
            c_out.append(c_1)
            h_out.append(h_1)
        self.block_stack.pop()
        save_names.restore()   # restore self.names
        return "".join(c_out + end_comment) , "".join(h_out)


    def do_join(self, args, quiet, source):
        """built-in function JOIN"""
        self.join_pending =  eval_codegen_join(self.names, args)
        return "", ""



    def parse_block_if(self, args, source):
        """Parser for reading an IF block from the source

        This is a variant of method parse_block() that also processes
        the 'ELSE' statementd appropriately.

        The function returns a triple

            (prefix_lines, code_lines, postfix_lines)

        similar to method parse_block(). List 'code_lines' contains
        only the lines not excluded by IF or ELSE clauses. List
        'prefix_lines' contains the ELSE clause after which valid
        lines follow (if any).
        """ 
        kwd_stack = ["IF"]
        out_prefix = []
        out_source = []
        else_ok = True
        else_done = condition = bool(safe_eval(args, self.names))
        for l in source:
            if self.verbose:
                level = len(self.block_stack) + len(kwd_stack) 
                print("%d. %s" % (level, l), end = "")
            new_condition = condition
            m = self.m_kwd.match(l)
            if m:
                sub_kwd, _, args = m.groups()
                if sub_kwd in self.block_kwds:
                    kwd_stack.append(sub_kwd)
                elif sub_kwd == "END":
                    end_kwd = direct_single_arg(args, kwd = "END")
                    match_kwd = kwd_stack.pop()
                    if end_kwd != match_kwd:
                        kw = ("END " + end_kwd, match_kwd)
                        s =" % does not match % in codegen directive"
                        raise TypeError(s % kw) 
                    if len(kwd_stack) == 0:
                        if not else_done:
                            out_prefix = []
                        return out_prefix, out_source, [l]
                elif sub_kwd == "ELSE" and len(kwd_stack) == 1:
                    new_condition = False
                    if not else_ok:
                        raise TypeError("Misplaced ELSE in IF directive in codegen")
                    if not else_done:
                        out_prefix = [l]
                        args = args.format(*self.args, **self.names) 
                        new_condition, else_ok = eval_codegen_else(
                            self.names, args)
                        else_done = new_condition
            if condition and new_condition:
                out_source.append(l)
            condition = new_condition 
             
        raise TypeError("Missing END IF directive in codegen")


    def block_if(self, args, quiet, source):
        """built-in function IF"""
        c_out, source, end_comment = self.parse_block_if(args, source)
        if quiet:
            c_out, end_comment = [], []
        h_out = []
        self.block_stack.append("IF")
        for c_1, h_1 in self.iter_generate_lines(iter(source)):
            c_out.append(c_1)
            h_out.append(h_1)
        self.block_stack.pop()
        return "".join(c_out + end_comment) , "".join(h_out)


    def process_use_table(self, line):
        """Do pending work of built-in function USE_TABLE"""
        if not self.use_table_pending:
            return
        self.C_table_name, h_out = False, ""
        found = line.find("[")
        if found >= 0:
            l1 = line[:found].split(" ")
            self.C_table_name = l1[-1].strip()
            if self.use_table_pending > 1:
                found = line.find("=")
                if found:
                    h_out = "extern " + line[:found].strip() + ";\n"
                    self.C_table_export = True
        self.use_table_pending = 0
        return "", h_out

        

    def process_export(self, line):                
        """Do pending work of built-in function EXPORT"""
        if not self.export_pending:
            return
        h_out =  line[:-1] + ";\n"
        if self.pxd_export_pending:
            self.pxd_entries.append(line[:-1].strip())
        self.export_pending = False
        self.pxd_export_pending = False
        return "", h_out

              
       

    def out_auto_headlines(self):
        s = """{1}
// This {0} file has been created automatically. Do not edit!!!
{1}

"""
        cmt = "/" * 77
        return s.format("C", cmt), s.format("C header", cmt)

            

    def iter_generate_lines(self, source):
        """This iterator is the workhorse for method generate(line).

        It reads input from iterator source, and it yields pairs
        (c_line, h_line), where c_line is written to the generated
        .c file and h_line is written to the generated .h file.

        This method may is also be called by block processing directives
        corresponding e.g. to an IF or FOR directive. In case of a
        FOR directive, this method called once for each iteration 
        of the FOR loop.
        """
        self.adjust_names()
        self.export_pending = False      # set when EXPORT action pending
        self.pxd_export_pending = False  # set when EXPORT include .pxd file
        self.use_table_pending = False   # set when USE_TABLE action pending
                                         # 1 = just remember C name of table
                                         # 2 = also create prototype
        self.join_pending = False        # set when JOIN action is pending
        for l in source:
            if self.verbose:
                print("%d: %s" % (len(self.block_stack), l), end = "")
            m = self.m_kwd.match(l)
            self.current_line = l
            if m:
                fname, quiet, args = m.groups()
                if not quiet:
                    yield l, ""
                args = format_line(args, self.names, self.args) 
                c_out, h_out = self.directives[fname](args, quiet, source)
                self.adjust_names()      # a directive might mess up names
                yield c_out, h_out
            else:
                l = format_line(l, self.names, self.args,
                                           self.is_terminal)
                l = indent_subsequent_lines(l) 
                c_out = l if self.gen_c else "" 
                h_out = l if self.gen_h else ""
                yield c_out, h_out
                if self.export_pending:
                    yield self.process_export(l)
                if self.use_table_pending:
                    yield self.process_use_table(l)
                self.join_pending = False


    def iter_generate(self, source):
        """Iterator for generating .c and .h file form a source file.

        The function yield pairs of lines. The first line of each pair
        is to be written into the .c file and the second line is to
        be written into the .h file.

        Parameter source should be an iterator that yields the lines of
        the source file.
        """
        self.C_table_name = None            # C name of table pending
        self.C_table_export = False         # export C name of table pending if set
        self.gen_c = True                   
        self.gen_h = False
        self.exported_table_names.clear()
        self.table_size_ = 0
        self.pxd_entries = []

        for c_out, h_out in self.iter_generate_lines(source):
            yield c_out, h_out



 
    @staticmethod
    def print_file(text, file):
        if text and file:
            print(text, end = "", file = file) 



    def iter_source(self, source):
        """An iterator that yields the lines of the source file.

        The source file (or the source string) is given by parameter 
        *source*. Description of parameter *source* see method 
        generate().

        The lines yielded by this function are fed into method
        iter_generate() of this class.
        """
        self.source_name = None
        if isinstance(source, list):
            for i, f in enumerate(source):
                for l in self.iter_source(f):
                    yield(l)
                if i < len(source) - 1:
                    yield "\n"
        elif isinstance(source, file):
            self.gen_c = True                   
            self.gen_h = False
            for l in source:
                yield l
        else:
            self.gen_c = True                   
            self.gen_h = False
            try:
                is_filename = not "\n" in source
            except:
                is_filename = True
            if is_filename:
                if not self.source_name:
                    self.source_name = source
                if self.verbose:
                    print(" > Opening file %s" % source)
                with open(source, "rt") as f:
                    for l in f:
                        yield l
                if self.verbose:
                    print(" > Closing file %s" % source)
               
            else:
                if source[-1] == "\n":
                    source = source[:-1]
                for l in source.split("\n"):
                    yield l + "\n"     
                     

    def generate(self, source, c_filename, h_filename=None):
        r"""Generate .c file and .h file from source file

        :param source:

           The source file to processed for code generation.
           This may be

             * A string containing at least one *newline* character.
               Then that string is directly interpreted as the content
               of the source.

             * Any other string. 
               Then that string is interpreted as a name of a file

             * A list of strings.
               This is interpreted as a sequence of strings or file names
               according to the last two rules. Then the items given by
               the entries of that sequence are concatenated.      

        :param c_filename:

           Name of the .c file to be generated or ``None``. 
           The extension must be part of that name.
      
        :param h_filename:

           Name of the .h file to be generated or ``None``
           (default). 
           The extension must be part of that name.
      
        """
        inp = self.iter_source(source)         # input source file
        c_out, h_out = self.out_auto_headlines()
        if not c_filename or isinstance(c_filename, file):
            c_file = c_filename                # output .c file
        else:
            c_file = open(c_filename,"wt")     # output .c file
            self.print_file(c_out, c_file)
        if  not h_filename or isinstance(h_filename, file): 
            h_file = h_filename                # output .h file
        else:
            h_file = open(h_filename,"wt")     # output .h file
            self.print_file(h_out, h_file)

        for c_out, h_out in self.iter_generate(inp):
            self.print_file(c_out, c_file)
            self.print_file(h_out, h_file)
        inp = None

        if c_filename and not isinstance(c_filename, file):
            c_file.close()
        if h_filename and not isinstance(h_filename, file):
            h_file.close()

    def table_size(self):
        return self.table_size_


    def generate_pxd(self, pxd_file, h_file_name, source = None, nogil = False):
        """Create .pxd file 'pxd_file' from last call to generate().

        After calling method generate(), all exported functions 
        preceded by an ``EXPORT`` directive are collected automatically. 

        Method ``generate_pxd`` creates a .pxd file with name ``pxd_file`` 
        that contains information about these exported function.

        A .pxd file is used in Cython. Its content is::

           cdef extern from <h_file_name>:
               <exported function 1>
               <exported function 2>
               ...

        The name  of the .h file is given as parameter and the exported
        functions listed below that statement are those collected by
        the last recent call to method generate(). 
 
        :param pxd_file:

            Name of the output file.

        :param h_file_name:

            Name of the .h file to be placed in the
            ``cdef extern from`` statement.
    
        :param source:
     
             An optional string written into the output file
             in front of the ``cdef`` statement.  Format of parameter 
             *source* is the same as in method generate()

        :param nogil:

             Optional, exported functions are declared as ``nogil``
             when set.
        """
        s_gil = " nogil" if nogil else ""
        f = open(pxd_file, "wt")
        print("# This .pxd file has been generated automatically. Do not edit!\n",
            file = f)     
        if source:
            old_source_name = self.source_name
            for l in self.iter_source(source):
                print(l, file = f, end = "")
            print("", file = f)
            self.source_name = old_source_name
        print('cdef extern from "{0}"{1}:'.format(h_file_name, s_gil), file = f)  
        for l in self.pxd_entries:
            print("    " + l, file = f)
        if len(self.pxd_entries) == 0:
            print("    pass", file = f)
        f.close()








def make_doc(source_file, output_file, tables = None):
    r"""Extract documentation from C file, Deprecated!!

    This function extracts documentation from C file that has usually
    been generated by class ``TableGenerator``. The function is 
    deprecated, since future versions of this project will use doxygen.

    Here parameter ``source_file`` describes a source file to be used
    as input for an instance of class ``TableGenerator``. 

    Parameter ``source_file may`` be:
        * the name of a C source file
        * a string containing C statements with at least one *newline*
          character
        * a list of file names or string as defined above

    Parameter ``output_file`` is the name of an output file to be generated.
    The output file contains function declarations and comments for
    each in the same way as in the processed source file. The bodies
    of the C functions are dropped.

    The source file is processed essentially in the same way as a
    source file is processed by an instance of class  ``TableGenerator``.

    If one of the ``TableGenerator`` directives ``EXPORT`` or ``COMMENT``
    is found in the source file then everything of the source file is 
    copied into the  output file up to and including the first comment.
 
    Here a comment line must begin with ``'/*'`` (preceded by white
    space only) and end with ``'*/'``, or it must be as sequence of lines
    containing  ``'//'`` preceded by white  space only. 

    Parameter ``tables`` is an optional dictionary that plays the same 
    role as in the constructor of class ``TableGenerator``.
    """
    def iter_source(source_file):
        if isinstance(source_file, list):
            for i, f in enumerate(source_file):
                for l in self.iter_source(f):
                    yield(l)
                if i < len(source_file) - 1:
                    yield "\n"
        else:
            try:
                is_filename = not "\n" in source
            except:
                is_filename = True
            if is_filename:
                with open(source_file, "rt") as f:
                    for l in f:
                        yield l               
            else:
                if source_file[-1] == "\n":
                    source_file = source_file[:-1]
                for l in source_file.split("\n"):
                    yield l + "\n"     
    warnings.warn(
        "Function make_doc() is deprecated. One may use doxygen instead",
         DeprecationWarning
    )


    g = open(output_file, "wt")
    m_copy_comment =   re.compile(r"\s*//\s*\%\%(EXPORT|COMMENT)\b")
    m_begin_comment =  re.compile(r"\s*/\*.*(\*/)?")
    m_end_comment = re.compile(r".*\*/")
    m_is_comment = re.compile(r"\s*//")
    m_stop = re.compile(r"\s*[{$]")
    m_any_directive =   re.compile(r"\s*//\s*\%\%[A-Za-z0-9_]*\b")
    comment_pending = in_comment = copy = is_keyword = False 
    for l in iter_source(source_file):
        is_keyword = False
        if  m_copy_comment.match(l):
            comment_pending  = is_keyword = True
            in_comment, copy = False, True
        elif m_begin_comment.match(l):
            if comment_pending:
                copy, comment_pending = True, False
                closing = m_begin_comment.match(l).groups()[0]
                in_comment = not closing
                comment_pending = False
        elif in_comment and m_end_comment.match(l):
            in_comment = False
        elif m_is_comment.match(l):
            if comment_pending:
                copy, comment_pending = True, False
        else:
            if not in_comment:
                if copy and not comment_pending:
                    print("\n\n",  file = g)
                    copy = False

        if  copy and not is_keyword:
            try:
                l = format_line(l, tables, terminal=0)
            except:
                pass 
            if not m_any_directive.match(l): 
                print(l, end = "", file = g)
    g.close()        
 



def c_snippet(source, *args, **kwds):
    r"""Return a C code snippet as a string from a *source* string

    Here ``source`` is a string that is interpreted in the same way 
    as the text in a source file in method ``generate()`` of class 
    ``TableGenerator``.
    The function applies the code generator to the string ``source``
    and returns the generated C code as a string.

    All subsequent keyword arguments are treated as a dictionary and
    they are passed to the code generator in class ``TableGenerator``
    in the same way as parameter ``tables`` in the constructor of 
    that class. 

    One can also pass positional arguments to this function. In the
    ``source`` string they an be accessed as ``%{0}``, ``%{1}``, etc.

    A line starting with ``// %%`` is interpreted as a directive as
    in class ``TableGenerator``.

    The keyword ``directives`` is reserved for passing a dictionary
    of directives as in class ``TableGenerator``.

    In order to achieve a similar effect as generating C code with::

         tg = TableGenerator(tables, directives)
         tg.generate(source_file, c_file)

    you may code e.g.::
        
        c_string = c_snippet(source, directives=directives, **tables) 

    If this function cannot evaluate an expression of shape ``%{xxx}`` 
    then the expression it is not changed; so a subsequent code 
    generation step may evaluate that expression. An unevaluated 
    argument in a directive leads to an error.
    """ 
    src = source.splitlines(True)
    try:
        directives = kwds['directives']
    except KeyError:
        directives = {}
    tg = TableGenerator(kwds, directives)
    tg.is_terminal = False
    tg.args = args
    return "".join(c for c,h in tg.iter_generate(src)) 






from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.structures.qs_matrix import QStateMatrix, qs_rand_matrix

from mmgroup.tests.test_clifford.test_qs_matrix import compare_complex


#####################################################################
# Test method rot_bits of class QStateMatrix
#####################################################################

rot_data = [
     (4, [0b000], 3, 2, 1),
     (4, [0b110_1001, 0b101_0111, 0b011_0100], 2, 1, 1),
]

def rot_testdata():
    for cols, data, rot, nrot, n0 in rot_data:
        yield QStateMatrix(0, cols, data), rot, nrot, n0
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                nrot = n0 = cols
                while n0 + nrot > cols:
                    n0, nrot = randint(0, cols), randint(0, cols)
                rot = randint(-nrot, nrot)
                m = qs_rand_matrix(0, cols, r)
                m.mul_scalar(randint(-8, 8), randint(0,7))  
                yield m, rot, nrot, n0

def rot_index(index, nqb, rot, nrot, n0):
    if  nrot <= 1: return index
    assert nrot + n0 <= nqb
    assert 0 <= index < 1 << nqb
    mask = ((1 << nrot) - 1) << n0
    i0, i1 = index & ~mask, index & mask
    rot %= nrot
    i1 = (i1 << rot) + (i1 >> (nrot - rot))
    return i0 + (i1 & mask)

       
@pytest.mark.qstate
def test_rot(verbose = 0):
    """Test function ``qstate12_rot_bits``. """
    for ntest, (m, rot, nrot, n0) in enumerate(rot_testdata()):
        if verbose:
            print("TEST rot bits %s" % (ntest+1))
            print(m)
            print("rot=", rot, ", nrot =", nrot, ", n0 =", n0)
        m1 = m.copy().rot_bits(rot, nrot, n0) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = np.copy(c)
        for i, x in enumerate(c):
            c1_ref[rot_index(i, m.ncols, rot, nrot, n0)] = x
        if verbose:
            print("output")
            print(m1)
        err = "Method rot has failed"
        compare_complex(c1_ref, c1, err)            
            
            
#####################################################################
# Test method rot_bits of class QStateMatrix
#####################################################################


xch_bits_data = [
     (4, [0b000], 2, 0b11),
     (4, [0b110_1001, 0b101_0111, 0b011_0100], 1, 0b101),
]

def xch_bits_testdata():
    for cols, data, sh, mask in xch_bits_data:
        yield QStateMatrix(0, cols, data), sh, mask
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                m = qs_rand_matrix(0, cols, r)
                m.mul_scalar(randint(-8, 8), randint(0,7))  
                if cols == 0:
                    yield m, 0, 0
                else:
                    sh = randint(0, cols - 1)
                    mask = randint(0, (1 << (cols - sh)) - 1)
                    mask &= ~(mask << sh)
                    yield m, sh, mask


def xch_bits_index(index, nqb, sh, mask):
    assert 0 <= index < 1 << nqb
    assert mask & (mask << sh) ==  0
    assert ((mask << sh) & (-1 << nqb)) == 0
    diff = mask & (index ^ (index >> sh))
    return index ^ diff ^ (diff << sh)


@pytest.mark.qstate
def test_xch_bits(verbose = 0):
    """Test function ``qstate12_xch_bits``. """
    for ntest, (m, sh, mask) in enumerate(xch_bits_testdata()):
        if verbose:
            print("TEST xch bits %s" % (ntest+1))
            print(m)
            print("sh=", sh, ", mask =", mask)
        m1 = m.copy().xch_bits(sh, mask) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = np.copy(c)
        for i, x in enumerate(c):
            c1_ref[xch_bits_index(i, m.ncols, sh, mask)] = x
        if verbose:
            print("output")
            print(m1)
        err = "Method xch_bits has failed"
        compare_complex(c1_ref, c1, err)            


#####################################################################
# Test methods extend and extend_zero
#####################################################################



extend_data = [
     (4, [0b000], 2, 2),
     (4, [0b110_1001, 0b101_0111, 0b011_0100], 1, 3),
]


def extend_testdata():
    for cols, data, j, nqb in extend_data:
        yield QStateMatrix(0, cols, data), j, nqb
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                m = qs_rand_matrix(0, cols, r)
                m.mul_scalar(randint(-8, 8), randint(0,7)) 
                j = randint(0, cols)
                nqb = randint(0, 5) 
                yield m, j, nqb



def extend_complex(c, j, nqb, zero = False):
    c1 = np.zeros(len(c) << nqb, dtype = complex)
    mask = (1 << j) - 1
    for i, x in enumerate(c):
        c1[(i & mask) + ((i & ~mask) << nqb)] = x
    if not zero:
        for i0 in range(0, len(c1), 1 << (nqb + j)):
            for i1 in range(1 << j):
                i = i0 + i1
                x = c1[i]
                c1[i : i + (1 << (nqb + j)) : 1 << j] = x
    return c1


@pytest.mark.qstate
def test_extend(verbose = 0):
    """Test function ``qstate12_extend``. """
    for ntest, (m, j, nqb) in enumerate(extend_testdata()):
        if verbose:
            print("TEST extend %s" % (ntest+1))
            print(m)
            print("j=", j, ", nqb =", nqb)
        m1 = m.copy().extend(j, nqb) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = extend_complex(c, j, nqb)
        if verbose:
            print("output after extend")
            print(m1)
        err = "Method extend has failed"
        compare_complex(c1_ref, c1, err)            

        m1 = m.copy().extend_zero(j, nqb) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = extend_complex(c, j, nqb, zero = True)
        if verbose:
            print("output after extend_zero")
            print(m1)
        err = "Method extend_zero has failed"
        compare_complex(c1_ref, c1, err)            


#####################################################################
# Test methods restrict, restrict_zero and sumup
#####################################################################


restrict_data = [
     (4, [0b0000], 2, 2),
     (4, [0b110_1001, 0b101_0111, 0b011_0100], 1, 3),
]


def restrict_testdata():
    for cols, data, j, nqb in restrict_data:
        yield QStateMatrix(0, cols, data), j, nqb
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                m = qs_rand_matrix(0, cols, r)
                m.mul_scalar(randint(-8, 8), randint(0,7)) 
                j = randint(0, cols)
                nqb = randint(0, cols - j) 
                yield m, j, nqb


def restrict_complex(c, j, nqb, zero = False):
    if not zero:
        mask = (1 << j) - 1
        c1 = np.zeros(len(c) >> nqb, dtype = complex) 
        for i in range(len(c1)):
            c1[i] = c[(i & mask) + ((i & ~mask) << nqb)]
    else:
        mask = ((1 << nqb) - 1) << j
        c1 = np.copy(c)
        for i, x in enumerate(c1):
            if i & mask: 
                c1[i] = 0
    return c1


def sumup_complex(c, j, nqb):
    c1 = c.reshape(-1, 1 << nqb, 1 << j)
    c1 = np.einsum("ijk->ik", c1)
    return c1.reshape(-1)

@pytest.mark.qstate
def test_restrict(verbose = 0):
    """Test function ``qstate12_restrict``. """
    for ntest, (m, j, nqb) in enumerate(restrict_testdata()):
        if verbose:
            print("TEST restrict %s" % (ntest+1))
            print(m)
            print("j=", j, ", nqb =", nqb)
        m1 = m.copy().restrict(j, nqb) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = restrict_complex(c, j, nqb)
        if verbose:
            print("output after restrict")
            print(m1)
        err = "Method restrict has failed"
        compare_complex(c1_ref, c1, err)            

        m1 = m.copy().restrict_zero(j, nqb) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = restrict_complex(c, j, nqb, zero = True)
        if verbose:
            print("output after restrict_zero")
            print(m1)
        err = "Method restrict_zero has failed"
        compare_complex(c1_ref, c1, err)    


        m1 = m.copy().sumup(j, nqb) 
        c, c1 = m.complex().ravel(), m1.complex().ravel() 
        c1_ref = sumup_complex(c, j, nqb)
        if verbose:
            print("output after sumup")
            print(m1)
        err = "Method sumup has failed"
        compare_complex(c1_ref, c1, err)    

        



r"""We deal with the 196884-dimensional representation of the monster.

Here class ``MMSpaceCRT`` is an analogue of class ``MMSpace`` that
allows a limited set of operations on the real 
``196884``-dimensional representation of the monster with vectors of
a small norm. Typical examples of such vectors are unit vectors, or
so called *axes*, as described in :cite:`Con85`.

"""
# References in the __docstr__ see file docs/source/references.bib


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import re
import numpy as np
from numbers import Integral
import warnings
from collections import defaultdict, OrderedDict
import math 
from functools import partial




from mmgroup.structures.abstract_group import singleton
from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepVector
from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepSpace
from mmgroup.structures.mm_space_indices import tuple_to_sparse
from mmgroup.structures.mm_space_indices import numeric_index_to_sparse
from mmgroup.structures.mm_space_indices import sparse_from_indices
from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.mm_space  import MMSpace, MMVector, mm_ops
from mmgroup.mm_space  import standard_mm_group
from mmgroup.structures.parse_atoms import AtomDict 
from mmgroup.structures.parse_atoms import eval_atom_expression, ihex 

from mmgroup.mm import mm_vector, mm_aux_random_mmv
from mmgroup.mm import mm_aux_zero_mmv, mm_aux_reduce_mmv
from mmgroup.mm import mm_aux_mmv_to_sparse
from mmgroup.mm import mm_aux_mmv_set_sparse
from mmgroup.mm import mm_crt_combine, mm_crt_check_v2
from mmgroup.mm import mm_crt_combine_bytes
from mmgroup.mm import mm_crt_check_g
from mmgroup.mm import mm_crt_norm_int32
from mmgroup.generators import mm_group_n_clear
from mmgroup.generators import mm_group_n_mul_word_scan
from mmgroup.generators import mm_group_n_to_word




_mm_op, _mm_compare = {},  {}
for p in (7, 31, 127, 255):
    _mm_op[p] = mm_ops[p].op_word
    _mm_compare[p] = mm_ops[p].op_compare

PRECISION = math.log(7 * 31 * 127 * 255) / math.log(2.0) - 4



######################################################################
# Auxiliary class vsparse representing a sparse vector
######################################################################


ERR_CRT_TYPE = "Connot construct MMVectorCRT object from type '%s'"


class vsparse:
    def __init__(self, *data):
        self.d = defaultdict(int)
        if len(data) == 0 or not data[0]:
            return
        scalar = 1
        if isinstance(data[0], vsparse):
            self.d.update(data[0].d)
            return
        if isinstance(data[0], Integral):
            scalar, data = data[0],  data[1:]
        if scalar == 0 or len(data) == 0:
            return
        if isinstance(data[0], str):
            if len(data[0]) != 1 or not data[0] in "ABCTXZYIDE0":
                err = "Illegal tag '%s' in tuple for class MMSpaceCRT"
                raise ValueError(err %  data[0]) 
        else:
            raise TypeError(ERR_CRT_TYPE % type(data[0]))          
        for t in tuple_to_sparse(255, *data):
            scalar1, t =  t & 0xff, t & 0xffffff00
            scalar1 = scalar1 if scalar1 < 128 else scalar1 - 255
            self.d[t] += scalar * scalar1

    def __imul__(self, other):
        assert isinstance(other, Integral)
        if other == 0:
            self.d.clear()
        else:
            for tag in self.d.keys():
                 self.d[tag] *= other
        return self

    def __mul__(self, other):
        return vsparse(self).__imul__(other)

    __rmul__ = __mul__

    def __neg__(self):
        return self.__mul__(-1) 

    def __pos__(self):
        return self  
 
    def __iadd__(self, other):
        if other == 0:
            return
        assert isinstance(other, vsparse)
        for tag, value in other.d.items():
            self.d[tag] += value
        return self

    def __add__(self, other):
        return vsparse(self).__iadd__(other)

    def __isub__(self, other):
        if other == 0:
            return
        assert isinstance(other, vsparse)
        for tag, value in other.d.items():
            self.d[tag] -= value
        return self

    def __sub__(self, other):
        return vsparse(self).__isub__(other)

    def reduce(self):
        for tag, value in self.d.items():
            if value == 0:
                del self.d[tag]
        return self
        
    def norm(self):
        norm = 0
        self.reduce()
        for tag, value in self.d.items():
            factor = 1
            if tag & 0xE000000 == 0x2000000:
                i0, i1 = (tag >> 14) & 0x7ff, (tag >> 8) & 0x3f
                factor += i0 != i1
            norm += factor * value * value
        return norm    

    def sparse_array(self, p, shift = 0):
        assert p & 1 and p < 256
        self.reduce()
        a = np.zeros(len(self.d), dtype = np.uint32)
        for i, (tag, value) in enumerate(self.d.items()):
            a[i] = tag + (value << shift) % p 
        return a    
  

######################################################################
# Convert string to instance of class vsparse
######################################################################
 
FRAME = re.compile(r"^([A-Za-z_])+\<([0-9]+;)?(.+)\>$") 


def vsparse_from_str(s):
    string = s
    m = FRAME.match(s)
    if m:
        _, p_str, string = m[1], m[2], m[3]
        if p_str:
            err = "Vector is defined module an integer only"
            raise ValueError(err)
    f = AtomDict(vsparse)
    return eval_atom_expression(string, f)
    
 

######################################################################
# Auxiliary functions
######################################################################



def _err_tag(*args, **kwds):
    err = "Bad entry in monster group element"
    raise ValueError(err)



def _iter_group(g):
    start, data, len_ = 0, g.mmdata, len(g.mmdata)
    while start < len_:
        tag = data[start] & 0x70000000
        if  tag >= 0x50000000:
            yield  data[start : start+1], True
            start += 1
        else:
            end_ = start
            while end_ < len_ and data[end_] & 0x70000000 < 0x50000000:
               end_ += 1
            yield data[start : end_], False
            start = end_ 


def compress_data(data):
    out = np.zeros(196884, dtype = np.int32)
    out[0:24] = data[0:768:33]
    k = 24
    for i in range(1,24):
        out[k:k+i] = data[32*i:32*i+i]
        out[276+k:276+k+i] = data[768+32*i:768+32*i+i]
        out[552+k:552+k+i] = data[1536+32*i:1536+32*i+i]
        k += i
    assert k == 300
    out[852:49428] = data[2304:50880]
    t = data[50880:247488].reshape((3*2048,32))[:,:24]
    out[49428:196884] = t.ravel()
    assert len(out) == 196884
    return out
    

######################################################################
# If check_MMVectorCRT is True then we check a vector of type
# MMVectorCRT for overflow or underflow after each operation
######################################################################

check_MMVectorCRT = True 

######################################################################
# Modelling a vector of the 196884-dimensional rep of the monster
######################################################################

MODULUS = 7*31*125*255
MAX_CRT_NORM = MODULUS**2 // 4
ERR_OVERFLOW = "Overflow in class MMVectorCRT"
ERR_UNDERFLOW = "Underflow in class MMVectorCRT"
ERR_UNDERFLOW_G = "Underflow at group operation in class MMVectorCRT"



class MMVectorCRT(AbstractMmRepVector):
    """Models a vector in a space of type ``MMSpaceCRT``.

    Such a vector should be constructed by calling an instance ``V``
    of class ``MMSpaceCRT`` which models a real representation of
    the monster group. Calculations in this space are exact in
    fixed-point arithmetic with a precsion of about %.2f bits.    

    ValueError is raised in case of overflow or underflow.
    
    The functionality of this class is a subset of the functionality 
    of class ``MMVector``. See class ``MMSpaceCRT`` for details.

    A vector may also be reduced modulo ``p = 7, 31, 127, 255`` with
    the modulo operator ``%%``. Then the result is a vector in the 
    vector space ``MMSpace(p)``.
     
    :var space:
        This attribute contains the space to which the vector
        belongs. That space is an instance of class |MMSpaceCRT|.

    .. warning::
       The constructor of this class is not for public use! You
       may call an instance ``V`` of class  |MMSpaceCRT| for
       constructing vectors in the real representation space
       ``V`` of the monster group.
    """ % PRECISION
    p = MODULUS
    MAX_CRT_NORM = MAX_CRT_NORM

    def __init__(self, shift, tag = 0, i0 = None, i1 = None):
        self.shift = shift
        self.factor = 1.0 / (1 << shift)
        if not 3 <= self.shift <= 21:
            raise ValueError("Bad shift factor for class MMVectorCRT") 
        self.data = OrderedDict()
        self.data_int = np.zeros(247488, dtype = np.int32)
        self.expanded = False
        d = vsparse()
        if isinstance(tag, MMVectorCRT): 
            for p in (7, 31, 127, 255):
                self.data[p] = tag.data[p].copy()
            self.shl(self.shift - tag.shift)
            return
        elif isinstance(tag, Integral) and not tag:
            for p in (7, 31, 127, 255):
                self.data[p] = MMVector(p)
            return
        elif isinstance(tag, str) and len(tag) == 1:
            d += vsparse(tag, i0, i1)
        elif isinstance(tag, str):
            d += vsparse_from_str(tag)
        elif isinstance(tag, list):
            for x in tag:
                if isinstance(x,tuple):
                    d += vsparse(*x)
                elif isinstance(x, str):
                    d += vsparse_from_str(x)
        else:
            err = "Connot construct MMVectorCRT object from type '%s'"
            raise ValueError(err % type(tag))
        for p in (7, 31, 127, 255):
            self.data[p] = v = MMVector(p)
            ind = d.sparse_array(p, shift)
            mm_aux_mmv_set_sparse(p, v.data, ind, len(ind)) 
        if check_MMVectorCRT:
            self.expand()
            if self._inorm > MAX_CRT_NORM:
                raise ValueError(ERR_OVERFLOW)

    def expand(self):
        """Expand data of vector ``v`` with CRT

        In the array ``v.data_int`` the ``i``-th entry is computed 
        form the ``i``-th entries of the vector modulo 7, 31, 127,
        and 255. 
        """
        if not self.expanded:
            v2 = mm_crt_combine(self.data[7].data, 
                self.data[31].data, self.data[127].data, 
                self.data[255].data, self.data_int)  
            self._v2 = v2 - self.shift if v2 < 24 else 24        
            self._inorm = mm_crt_norm_int32(self.data_int)
            self.expanded = True


    def shl(self, sh):
        assert isinstance(sh, Integral)
        if sh == 0:
            return self
        if sh > 0:
            if check_MMVectorCRT:
                self.expand()
                if sh > 24 or self._inorm * 4**sh > MAX_CRT_NORM:
                    raise ValueError(ERR_OVERFLOW)
                self.data_int <<= sh  
                self._inorm *= 4**sh            
            for d in self.data.values:
                d <<= sh
            self._v2 += sh            
        elif sh < 0:
            nsh = -sh
            if check_MMVectorCRT:
                self.expand()
                if nsh > self.v2:
                    raise ValueError(ERR_UNDERFLOW)
                self.data_int >>= nsh           
                self._inorm *= 4**sh            
            for d in self.data.values:
                d >>= nsh
            self._v2 += sh            
        return self     

    def check(self):
        """Check if the vector is correct

        Raise ValueError if the vector is errorneous.
        """
        return True


    def __ilshift__(self, other):
        return self.shl(other)

    def __lshift__(self, other):
        return self.copy().shl(other)

    def __irshift__(self, other):
        return self.shl(-other)

    def __lrhift__(self, other):
        return self.copy().shil(-other)


    def __mod__(self, p):
        """Return the vector modulo ``p``. 

        ``p``  must be in (7, 31, 127, 255). Actually, we divide 
        the vector by the *scaling factor* befor reducing it 
        modulo ``p``.

        This method is mainly for testing.
        """ 
        if  p in (7, 31, 127, 255):
            return MMVector(p, self.data[p]) >> self.shift
        elif p in (3, 15):
            v0 = self % 255
            return MMVector(p, v0)
        elif isinstance(p, Integral):
            err = "Cannot reduce MMVectorCRT object modulo %d"
            raise ValueError(err % p)
        else:
            err = "Modulus for reducing MMVectorCRT object must be int"
            raise TypeError(err)

    @property
    def v2(self): 
        """Return the ``2``-adic value of the vector.
 
        If "a, b" are odd integers and ``k`` is an integer then
        the  ``2``-adic value  ``a * 2**k / b`` is ``k``.

        The  ``2``-adic value of a vector is the minimum of the
         ``2``-adic values of its entries, ignoring zero entries.

        The function raises ZeroDivisionError if ``v ==  0``
        """
        v2 = self._v2
        if self._v2 >= 24:
            err = "The zero vector has infinite 2-adic value"
            raise ZeroDivisionError(err)
        return self._v2
 
    @property
    def inorm(self):
        """Return a scaled norm of the vector as an integer

        For a vector ``v`` we have 

            ``v.fnorm() = v.inorm() * v.factor**2``.

        Where ``v.fnorm()`` is the real norm of ``v``.
        """
        self.expand()
        return self._inorm



    def norm(self):
        """Return norm of vector as a floating point number.

        The norm of a vector in the representation of the monster
        is the squared sum of is entries. Here the squares of all
        entries with index ``("A", i0, i1)`` must be doubled in
        case ``i0 != i1``.
 
        The returned norm is exact.
        """
        return self.inorm * self.factor**2

######################################################################
# class MMSpace
######################################################################


@singleton
class MMSpaceCRT(AbstractMmRepSpace):
    """Models a ``196884``-dimensional representation of the monster group 

    This class models a real representation of the monster group 
    with fixed-point arithmetic. Calcuations are done by combining
    vectors modulo ``p = 7, 31, 127, 255`` with Chinese remaindering.
    This way we achive about %.2f bit precision.

    The construction of a vector in this space and the computation
    with such vectors works in the same way as in class |MMSpace|.
    But there are som limitaions:

      * Vectors may be constructed as in class |MMSpace|, but
        the arguments of the constructor may be tuples only. 
        Here randomized scalars are illegal.

      * The only operations allowed for vectors are copying, 
        multiplication with a group element, and testing for 
        equality. Vector addition and scalar multiplication
        are illegal.

      * A vector may be reduced modulo one of the primes
        ``p = 7, 31, 127, 255`` using the modulo operatior ``%%``.
        The result is a vector in the space ``MMSpace(p)``.

      * Changing entries of a vector via item assignment is
        illegal. 

      * Entries and subarrays of a vector may be obtained as
        in class |MMSpace|. Here subarrays are returned as
        integers or numpy  arrays with ``dtype = np.int32``.
        On output, each number is multiplied with ``2**k``
        in order to obtain an integer. Here ``k`` is the
        argument given in the cnstructor by parameter ``shift``.
        
        You may multiply such an integer value with the attribute
        ``factor`` of the space for obtaining an exact floating
        point value. But floating point data may lead to imprecise
        results when processing the with a computer algebra system.

      
    :var shift:
        If shift is ``k`` then a vector ``v`` can be represented if 
        all entries of ``v`` are integer multiples of ``2**(-k)``.  
        The absolute value of an entry may be at most 
        ``3513772 / 2**k``. In case ``shift == 20`` (default)
        this is about  ``3.35``.
 
    :var group:
        Contains the group operating on the vector
        space by right multiplication. 
        That group must be an instance of class |MMGroup|.

    Caution!
    
        The norm of an input vector must not exceed 
        ``1.234 * 10**13 * 4.0 ** (-k)``. Otherwise *overflow*
        occurs. In case ``k = 20`` (default) this quantity is about 
        ``11.229``. Here ``k`` is given by parameter ``shift``. 

        The norm of a vector is the sum of the squares of its 
        entries, where for all entries with index ``("A", i0, i1)`` ,
        ``i0 != i1``, the doubled square of the entry must be 
        taken instead.

    Caution!

        When multiplying a vector ``v`` with a generator of the group 
        with tag ``t`` or ``l``, and ``v`` has an entry that is not 
        an integer multiple of ``8 * 2**(-k)``, then a 
        *precision error* may occur. We raise ``ValueError`` in 
        case of a precision error.
      

    """ % PRECISION
    vector_type = MMVectorCRT
    _max_norm = (7*31*125*255)**2 // 4

    def __init__(self):
        """Create a 196884-dimensional representation of the monster

        All calculations are done modulo the odd number p
        """
        super(MMSpaceCRT, self).__init__()

    #######################################################################
    # Conversion to a list of tuples 
    #######################################################################

    def _not_supported(self, *args, **kwds):
        err = "Method not supported in class MMSpaceCRT"
        raise NotImplementedError(err)

    as_tuples = _not_supported



    #######################################################################
    # Creating vectors 
    #######################################################################

    def zero(self, shift):
        """Return the zero vector"""
        err = "Cannot create zero vector in space of class MMSpaceCRT"
        return MMVectorCRT(shift)

    def copy_vector(self, v1):
        assert v1.space == self
        v = MMVectorCRT(v1.shift, 0)
        for p in (7, 31, 127, 255):
            np.copyto(v.data[p].data, v1.data[p].data)
        if v1.expanded:
            np.copyto(v.data_int, v1.data_int)
            v._v2 = v1._v2        
            v._inorm = v1._inorm
        v.expanded = v1.expanded
        return v

      
    def __call__(self, *args):
       return self.from_tuples(*args)      


    def set_rand_uniform(self, *args, **kwds):
        err = "Cannot create random vector in space of class MMSpaceCRT"
        raise NotImplementedError(err)


    parse = _not_supported

    #######################################################################
    # Obtaining and setting components via sparse vectors
    #######################################################################


    def getitems_sparse(self, v, a_sparse):
        raise NotImplementedError(err)

    def additems_sparse(self, *args, **kwds):
        err = "Sparse representation not supported in class MMSpaceCRT"
        raise NotImplementedError(err)

    setitems_sparse = additems_sparse


    #######################################################################
    # Conversion from and to to sparse representation 
    #######################################################################

    as_sparse = additems_sparse


    #######################################################################
    # Vector operations 
    #######################################################################


    def iadd(self, v1, v2):
        err = "Vector addition not supported in class MMSpaceCRT"
        raise NotImplementedError(err)
 
    def imul_scalar(self, v1, a):
        err = "Scalar multiplication not supported in class MMSpaceCRT"
        raise NotImplementedError(err)
           
    #######################################################################
    # Group operation 
    #######################################################################

    def _imul_word(self, v1, g_word, buf):
        if len(g_word):
            vectors = list(v1.data.values())
            if check_MMVectorCRT:
                if mm_crt_check_g(g_word[0], *[v.data for v in vectors]):
                    print("MMVectorCRT: shift = %d, v2 = %d, tag = %s" %
                        (v1.shift, v1.v2, ((g_word[0] >> 28) & 7)))
                    raise ValueError(ERR_UNDERFLOW_G)
            for v in vectors:
                v.ops.op_word(v.data, g_word, len(g_word), 1, buf)
        

    def imul_group_word(self, v1, g):
        """Return product v1 * g of vector v1 and group word g.

        v1 is replaced by v1 * g.

        This method is called for elements v1 of the space
        'self' and for elements g of the group 'self.group' only.
        """
        assert isinstance(g, AbstractGroupWord) and g.group.is_mmgroup 
        a = g.mmdata
        nn = np.zeros(5, dtype = np.uint32)
        nnw = np.zeros(5, dtype = np.uint32)
        buf = np.zeros(v1.data[255].ops.MMV_INTS, dtype = np.uint64)
        while len(a):
            mm_group_n_clear(nn)
            i = mm_group_n_mul_word_scan(nn, a, len(a))
            length =  mm_group_n_to_word(nn, nnw)
            self._imul_word(v1, nnw[:length], buf)
            a = a[i:]
            if len(a):
                self._imul_word(v1, a[:1], buf)    
                a = a[1:]
        del buf 
        v1.expanded = False
        return v1       
        
            
 


    vector_mul_exp = _not_supported

    #######################################################################
    # Checking equality
    #######################################################################

    def equal_vectors(self, v1, v2):
        """Return True iff vectors v1 and v2 are equal 

        This method is called for elements v1 and v2 of the space
        'self' only.
        """
        v1.expand()
        v2.expand()
        v1_2, v2_2 = v1._v2, v2._v2
        if v1_2 == v2_2 == 24:
             return True   # then v1 = v1 = zero
        if v1_2 != v2_2:
             return False
        data1, data2 = v1.data_int, v2.data_int
        if v1.shift > v2.shift:
             data1 = data1 >> (v1.shift - v2.shift)
        if v2.shift > v1.shift:
             data2 = data2 >> (v2.shift - v1.shift)
        return (data1 == data2).all()

    #######################################################################
    # Conversion from and to byte format
    #######################################################################

    as_bytes =  _not_supported

    from_bytes = _not_supported
        
    #######################################################################
    #  Checking and reducing a vector
    #######################################################################

    def check(self, v1):
        return True
 
    def reduce(self, v1):
        return v1

    #######################################################################
    # getitem and setitem
    #######################################################################

    
    def vector_get_item(self, v, item):
        assert v.space == self
        if not isinstance(item, tuple):
            item = (item,) 
        shape, a_sparse = sparse_from_indices(255, *item)
        d = {}
        for p in (7, 31, 127, 255):
            vdata = v.data[p]
            asp = a_sparse.copy()
            vdata.space.getitems_sparse(vdata, asp)        
            d[p] = (asp & p).astype(np.uint8)
        l = len(d[7])
        a = np.zeros(l, dtype = np.int32)
        mm_crt_combine_bytes(d[7], d[31], d[127], d[255], l, a)
        af = np.array(a * v.factor,  dtype = float)
        return af.reshape(shape) if len(shape) else float(af)
        


    def vector_set_item(*args, **kwd):
        err = "Item assigment not supported in space of type MMSpaceCRT"
        raise NotImplementedError(err) 
 

    #######################################################################
    # Formatting a vector 
    #######################################################################


    def str_vector(self, v1):
        return "<vector in space of type MMSpaceCRT>" 

 
    #######################################################################
    # Properties
    #######################################################################

    @property
    def factor(self):
        """Factor of type ``float`` for vector entries.

        When reading entries from a vector then integers or arrays
        of integers are returned. Here each entry must be multiplied
        with this ``factor`` to obtain an exact floating point value
        of that entry.
        """
        return self._factor       



    @property
    def max_norm(self):
        """Maximum feasible norm of an input vector.

        The norm of a vector is the sum of the squares of its 
        entries, where for all entries with index ``("A", i0, i1)`` ,
        ``i0 != i1``, the doubled square of the entry must be 
        taken instead.
        """
        return (self._max_norm * self._factor)**2        




StdMMSpaceCRT = MMSpaceCRT()
MMVectorCRT.space = StdMMSpaceCRT


def MMV_CRT(shift):
    return partial(MMVectorCRT, shift)
